/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>

#include "zm_mem.h"
#include "zm_data_buffer_impl.h"

static void print_data(const ZM_DATA_BUFFER  * buf)
{
	/* Get data                      */
	const char  * data     = zm_data_buffer_get(buf);
	/* Get data length               */
	const int32_t data_len = (int32_t)zm_data_buffer_get_size(buf);

	fprintf(stdout, "Data: `%.*s`\n", data_len, data);
}

int main(int argc, char ** argv)
{
	ZM_DATA_BUFFER     buf;
	ZM_DATA_BUFFER     buf2;
	ZM_DATA_BUFFER   * buf3;

	char             * buf_data = NULL;

	/*
	 * zm_default_allocator - default memory allocattor for ZeroMIME library
	 *                        you can write own allocator; see examples/zm_mem_allocator.c
	 */
	zm_data_buffer_create_impl(&buf,  &zm_default_allocator);

	/* Basic operations ///////////////////////////////////////////////////////////////////  */
	/* Append asciz string  */
	zm_data_buffer_append_str(&buf, "Hello, ");
	/* Append some octets (binary data) */
	zm_data_buffer_append_data(&buf, "World", sizeof("World") - 1);
	/* Append one character */
	zm_data_buffer_append_char(&buf, '!');
	print_data(&buf);

	/* Object creation /////////////////////////////////////////////////////////////////////  */
	/* Create new EMPTY object */
	zm_data_buffer_clone(&buf2, &buf);

	/* And fill with data      */
	zm_data_buffer_append_str(&buf2, " Other object!");

	/* More complex example: Create new object and clone data from first object to new one */
	buf3 = zm_data_buffer_clone(NULL, &buf);
	/*
	 * IMPORTANT: zm_data_buffer_clone does not copy data!
	 *            This function only creates new object with same memory allocator.
	 *            As you can see the clone has yet no data.
	 */
	print_data(buf3);
	/* Now we can copy our data */
	zm_data_buffer_assign(buf3, &buf);
	print_data(buf3);

	/* Append second object to the first */
	zm_data_buffer_append(&buf, &buf2);
	print_data(&buf);
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));

	/* Reserve some data */
	zm_data_buffer_reserve(&buf2, 1024);
	/* Yes, it's right: we're reserving 1024 octets of *free* space, so new size is 14 + 1024 = 1038 octets */
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));
	zm_data_buffer_reserve(&buf2, 1024);
	/* Object unchanged */
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));

	/* Reset object */
	zm_data_buffer_reset(&buf2);
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));
	print_data(&buf2);

	/*
	 * You can copy data directly to the object.
	 * But with great power comes great responsibility.
	 * You *must* check available space and set new size by hand
	 */

	/* We assume YOU know what you do */
	buf_data = zm_data_buffer_get(&buf2);
	memcpy(buf_data, "Directly copied data", sizeof("Directly copied data") - 1);
	/* And set new object size */
	zm_data_buffer_set_size(&buf2, sizeof("Directly copied data") - 1);
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));
	print_data(&buf2);

	/* You can copy data directly to the first unuset octet of the object */
	buf_data = zm_data_buffer_get_tail(&buf2);
	memcpy(buf_data, "; Part2.", sizeof("; Part2.") - 1);
	zm_data_buffer_set_size(&buf2, zm_data_buffer_get_size(&buf2) + sizeof("; Part2.") - 1);
	fprintf(stdout, "Size, unused size and capacity: %i %i %i\n", (int32_t)zm_data_buffer_get_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_tail_size(&buf2),
	                                                              (int32_t)zm_data_buffer_get_capacity(&buf2));
	print_data(&buf2);

	/* Release memory */
	zm_data_buffer_destroy(buf3);
	zm_data_buffer_destroy(&buf2);
	zm_data_buffer_destroy(&buf);


return EX_OK;
}
/* End. */
