/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_DATA_BUFFER_H__
#define _ZM_DATA_BUFFER_H__ 1

/**
 * @file zm_data_buffer.h
 * @brief Iris data buffer, base interface
 */

#include "zm_data_buffer_base_defs.h"
#include "zm_data_buffer_defs.h"
#include "zm_mem_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @struct _zm_data_buffer zm_data_buffer.h <zm_data_buffer.h>
 * @brief Data buffer
 */
struct ZM_API _zm_data_buffer
{
	/** Opaque data                        */
	void               * opaque;
	/** Data buffer, virtual methods table */
	ZM_DATA_BUFFER_VT  * vt;
	/* Flags                               */
	uint32_t             flags;
};

/**
 * @brief Create data buffer using existing object type and allocator
 * @param data_buffer - Destination object
 * @param data_buffer_src - Source object
 * @return Pointer to the data buffer
 */
ZM_DATA_BUFFER * zm_data_buffer_clone(ZM_DATA_BUFFER  * data_buffer,
                                      ZM_DATA_BUFFER  * data_buffer_src);

/**
 * @brief Destroy data buffer
 * @param data_buffer - Data buffer object
 */
void zm_data_buffer_destroy(ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Get allocator
 * @param data_buffer - Data buffer object
 * @return Pointer to the memory allocator
 */
ZM_ALLOCATOR * zm_data_buffer_get_allocator(ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Reset data buffer
 * @param data_buffer - Data buffer object
 */
void zm_data_buffer_reset(ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Get data
 * @param data_buffer - Data buffer object
 * @return Pointer to the data buffer
 */
char  * zm_data_buffer_get(const ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Get data length
 * @param data_buffer - Data buffer object
 * @return Data buffer size
 */
int64_t zm_data_buffer_get_size(const ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Get object capacity
 * @param data_buffer - Data buffer object
 * @return Data buffer capacity
 */
int64_t zm_data_buffer_get_capacity(const ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Append one character
 * @param data_buffer - Data buffer object
 * @param data - Character to add
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_append_char(ZM_DATA_BUFFER  * data_buffer,
                                   char              data);

/**
 * @brief Append data, null-terminated string
 * @param data_buffer - Data buffer object
 * @param data - Data to add, null-terminated string
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_append_str(ZM_DATA_BUFFER  * data_buffer,
                                  const char      * data);

/**
 * @brief Append data
 * @param data_buffer - Data buffer object
 * @param data - Data to add
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_append_data(ZM_DATA_BUFFER  * data_buffer,
                                   const void      * data,
                                   uint64_t          size);

/**
 * @brief Append other data buffer
 * @param data_buffer - Data buffer object
 * @param data - Data buffer to add
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_append(ZM_DATA_BUFFER        * data_buffer,
                              const ZM_DATA_BUFFER  * data);

/**
 * @brief Assign one character
 * @param data_buffer - Data buffer object
 * @param data - Data to assign, character
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_assign_char(ZM_DATA_BUFFER  * data_buffer,
                                   char              data);

/**
 * @brief Assign data, null-terminated string
 * @param data_buffer - Data buffer object
 * @param data - Data to assign, null-terminated string
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_assign_str(ZM_DATA_BUFFER  * data_buffer,
                                  const char      * data);

/**
 * @brief Assign data
 * @param data_buffer - Data buffer object
 * @param data - Data to assign
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_assign_data(ZM_DATA_BUFFER  * data_buffer,
                                   const void      * data,
                                   uint64_t          size);

/**
 * @brief Assign other data buffer
 * @param data_buffer - Data buffer object
 * @param data - Data buffer to assign
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_assign(ZM_DATA_BUFFER        * data_buffer,
                              const ZM_DATA_BUFFER  * data);

/**
 * @brief Reserve space
 * @param data_buffer - Data buffer object
 * @param size - New size
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_reserve(ZM_DATA_BUFFER  * data_buffer,
                               uint64_t          size);

/**
 * @brief Set size
 * @param data_buffer - Data buffer object
 * @return New data size or -1 if any error occured
 */
int64_t zm_data_buffer_set_size(ZM_DATA_BUFFER  * data_buffer,
                                uint64_t          size);

/**
 * @brief Get free space length (capacity - size)
 * @param data_buffer - Data buffer object
 * @return Free space length
 */
int64_t zm_data_buffer_get_tail_size(const ZM_DATA_BUFFER  * data_buffer);

/**
 * @brief Get pointer to the tail of buffer, started at current size
 * @param data_buffer - Data buffer object
 * @return Pointer to data
 */
char * zm_data_buffer_get_tail(const ZM_DATA_BUFFER  * data_buffer);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_DATA_BUFFER_H__ */
/* End */
