/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_DATA_BUFFER_BASE_H__
#define _ZM_DATA_BUFFER_BASE_H__ 1

/**
 * @file zm_data_buffer_base.h
 * @brief Iris data buffer, base interface
 */

#include "zm_data_buffer_base_defs.h"
#include "zm_mem_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief Create data buffer function
 * @param alloc - Memory allocator
 * @return Pointer to the opaque data for newly created data buffer
 */
typedef void * (*zm_data_buffer_create_fn)(ZM_ALLOCATOR  * alloc);

/**
 * @brief Destroy data buffer function
 * @param opaque - Opaque data for base object
 */
typedef void (*zm_data_buffer_destroy_fn)(void  * opaque);

/**
 * @brief Get allocator function
 * @param opaque - Opaque data for base object
 * @return Pointer to the memory allocator
 */
typedef ZM_ALLOCATOR * (*zm_data_buffer_get_allocator_fn)(void  * opaque);

/**
 * @brief Reset data buffer function
 * @param opaque - Opaque data for base object
 */
typedef void (*zm_data_buffer_reset_fn)(void  * opaque);

/**
 * @brief Get data function
 * @param opaque - Opaque data for base object
 * @return Pointer to the data
 */
typedef char * (*zm_data_buffer_get_fn)(const void  * opaque);

/**
 * @brief Get data length function
 * @param opaque - Opaque data for base object
 * @return Data size
 */
typedef int64_t (*zm_data_buffer_get_size_fn)(const void  * opaque);

/**
 * @brief Append one character function
 * @param opaque - Opaque data for base object
 * @param data - Character to add
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_append_char_fn)(void  * opaque,
                                                 char    data);

/**
 * @brief Append data function
 * @param opaque - Opaque data for base object
 * @param data - Data to add
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_append_data_fn)(void        * opaque,
                                                 const char  * data,
                                                 uint64_t      size);

/**
 * @brief Append data function
 * @param opaque - Opaque data for base object
 * @param data - Data to add, null-terminated string
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_append_str_fn)(void        * opaque,
                                                const char  * data);

/**
 * @brief Assign data function
 * @param opaque - Opaque data for base object
 * @param data - Data to add
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_assign_data_fn)(void        * opaque,
                                                 const char  * data,
                                                 uint64_t      size);

/**
 * @brief Assign data function
 * @param opaque - Opaque data for base object
 * @param data - Data to add, null-terminated string
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_assign_str_fn)(void        * opaque,
                                                const char  * data);

/**
 * @brief Reserve space function
 * @param opaque - Opaque data for base object
 * @param size - New size
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_reserve_fn)(void      * opaque,
                                             uint64_t    size);

/**
 * @brief Set size function
 * @param opaque - Opaque data for base object
 * @param size - New size
 * @return New data size or -1 if any error occured
 */
typedef int64_t (*zm_data_buffer_set_size_fn)(void      * opaque,
                                              uint64_t    size);

/**
 * @brief Get free space length function
 * @param opaque - Opaque data for base object
 * @return Free space length
 */
typedef int64_t (*zm_data_buffer_get_tail_size_fn)(void  * opaque);

/**
 * @brief Get pointer to the tail of buffer, started at current size
 * @param opaque - Opaque data for base object
 * @return Pointer to data
 */
typedef char * (*zm_data_buffer_get_tail_fn)(void  * opaque);

/**
 * @struct _zm_data_buffer zm_data_buffer.h <zm_data_buffer.h>
 * @brief Data buffer, virtual methods table
 */
struct ZM_API _zm_data_buffer_vt
{
	/** Create data buffer              */
	zm_data_buffer_create_fn         create_fn;
	/** Destroy data buffer             */
	zm_data_buffer_destroy_fn        destroy_fn;
	/** Get allocator                   */
	zm_data_buffer_get_allocator_fn  get_allocator_fn;
	/** Reset data buffer               */
	zm_data_buffer_reset_fn          reset_fn;
	/** Get data                        */
	zm_data_buffer_get_fn            get_fn;
	/** Get data length                 */
	zm_data_buffer_get_size_fn       size_fn;
	/** Append data                     */
	zm_data_buffer_append_data_fn    append_data_fn;
	/** Assign data                     */
	zm_data_buffer_assign_data_fn    assign_data_fn;
	/** Reserve space                   */
	zm_data_buffer_reserve_fn        reserve_fn;
	/** Set size                        */
	zm_data_buffer_set_size_fn       set_size_fn;
	/** Get free space length           */
	zm_data_buffer_get_tail_size_fn  get_tail_size_fn;
	/** Get pointer                     */
	zm_data_buffer_get_tail_fn       get_tail_fn;
};

#ifdef __cplusplus
}
#endif

#endif /* _ZM_DATA_BUFFER_BASE_H__ */
/* End */
