/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_ICONV_H__
#define _ZM_ICONV_H__ 1

/**
  @file zm_inconv.h
  @brief Zero MIME encoding converter
*/

#include "zm_data_buffer_defs.h"
#include "zm_iconv_defs.h"
#include "zm_mem_defs.h"
#include "zm_rbtree.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @struct _zm_iconv zm_iconv.h <zm_iconv.h>
 * @brief Text converter
 */
struct ZM_API _zm_iconv
{
	/** Map of encodings */
	ZM_RB_TREE      encodings_map;

	/** Opaque data allocator */
	ZM_ALLOCATOR  * alloc;

	/* Flags */
	uint32_t        flags;
};

/**
 * @brief Create text converter
 * @param zm_iconv - Iconv object
 * @param alloc - Memory allocator
 * @return Pointer to the data buffer
 */
ZM_ICONV * zm_iconv_create(ZM_ICONV      * zm_iconv,
                           ZM_ALLOCATOR  * alloc);

/*
 * @brief Add encoding alias
 * @param zm_iconv - Iconv object
 * @param encoding - Encoding, asciz string
 * @param alias - Encoding alias, asciz string
 * @return 0 - if success, -1 - otherwise
 */
int32_t zm_iconv_add_alias(ZM_ICONV    * zm_iconv,
                           const char  * encoding,
                           const char  * alias);

/*
 * @brief Remove encoding alias
 * @param zm_iconv - Iconv object
 * @param alias - Encoding alias, asciz string
 * @return 0 - if success, -1 - otherwise
 */
int32_t zm_iconv_remove_alias(ZM_ICONV    * zm_iconv,
                              const char  * alias);

/**
 * @brief Convert text
 * @param zm_iconv - Iconv object
 * @param src - Source data
 * @param src_len - Source data length
 * @param result - Result data [out]
 * @param result_len - Result data length [out]
 * @param src_charset - Source character set
 * @param dst_charset - Destination character set
 * @param errors - Number of errors
 * @param ignore_ilseq - If set to 1, ignore illegal sequences
 * @return 0 - if success, -1 - otherwise
 */
int32_t zm_iconv_convert(ZM_ICONV        * zm_iconv,
                         const void      * src,
                         const uint64_t    src_len,
                         void           ** result,
                         uint64_t        * result_len,
                         const char      * src_charset,
                         const char      * dst_charset,
                         uint32_t        * errors,
                         const int32_t     ignore_ilseq);

/**
 * @brief Convert text
 * @param zm_iconv - Iconv object
 * @param src - Source data
 * @param src_len - Source data length
 * @param result - Result data buffer [out]
 * @param src_charset - Source character set
 * @param dst_charset - Destination character set
 * @param errors - Number of errors
 * @param ignore_ilseq - If set to 1, ignore illegal sequences
 * @return 0 - if success, -1 - otherwise
 */
int32_t zm_iconv_convert_data_buffer(ZM_ICONV        * zm_iconv,
                                     const void      * src,
                                     const uint64_t    src_len,
                                     ZM_DATA_BUFFER  * result,
                                     const char      * src_charset,
                                     const char      * dst_charset,
                                     uint32_t        * errors,
                                     const int32_t     ignore_ilseq);

/**
 * @brief Convert text
 * @param zm_iconv - Iconv object
 * @param src - Source data buffer
 * @param result - Result data buffer [out]
 * @param src_charset - Source character set
 * @param dst_charset - Destination character set
 * @param errors - Number of errors
 * @param ignore_ilseq - If set to 1, ignore illegal sequences
 * @return 0 - if success, -1 - otherwise
 */
int32_t zm_iconv_convert_data_buffer2(ZM_ICONV        * zm_iconv,
                                      ZM_DATA_BUFFER  * src,
                                      ZM_DATA_BUFFER  * result,
                                      const char      * src_charset,
                                      const char      * dst_charset,
                                      uint32_t        * errors,
                                      const int32_t     ignore_ilseq);

/**
 * @brief Destroy text converter
 * @param zm_iconv - Iconv object
 */
void zm_iconv_destroy(ZM_ICONV  * zm_iconv);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_ICONV_H__ */
/* End. */
