/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_ITERATOR_H__
#define _ZM_ITERATOR_H__ 1

/**
  @file zm_iterator.h
  @brief Zero MIME data iterator
*/

#include "zm_iterator_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief MIME parser iterator
 */
struct ZM_API _zm_iterator
{
	/** Message body         */
	const void     * data;
	/** Current position     */
	uint64_t         pos;
#ifdef ZM_EXTRA_STREAM_INFO
	/** Max. stream bits     */
	unsigned char    bits;
	/** Line length          */
	uint64_t         line_len;
	/** Max. line length     */
	uint64_t         max_line_len;
#endif
};

/**
 * @brief Create iterator
 * @brief it - Iterator object [out]
 * @brief data - Source data block
 * @brief data - Iterator position in block
 */
void zm_iterator_create(ZM_ITERATOR  * it,
                        const void   * data,
                        uint64_t       pos);

/**
 * @brief Create empty iterator
 * @brief it - Iterator object [out]
 */
void zm_iterator_zero(ZM_ITERATOR  * it);

/**
 * @brief Copy iterator
 * @brief it_dst - Destination iterator object [out]
 * @brief it_src - Source iterator object
 */
void zm_iterator_copy(ZM_ITERATOR        * it_dst,
                      const ZM_ITERATOR  * it_src);

/**
 * @brief Compare iterators
 * @brief it_x - First iterator to compare
 * @brief it_y - Second iterator to compare
 * @return 0 - if equal, -1 - otherwise
 */
int32_t zm_iterator_equal(const ZM_ITERATOR  * it_x,
                          const ZM_ITERATOR  * it_y);

/**
 * @brief Compare iterators
 * @brief it_x - First iterator to compare
 * @brief it_y - Second iterator to compare
 * @return 0 - if NOT equal, -1 - otherwise
 */
int32_t zm_iterator_not_equal(const ZM_ITERATOR  * it_x,
                              const ZM_ITERATOR  * it_y);

/**
 * @brief Get current iterator position
 * @brief it - Iterator object
 * @return Iterator position
 */
uint64_t zm_iterator_get_pos(const ZM_ITERATOR  * it);

/**
 * @brief Get disatnce between two iterators
 * @brief it_start - First iterator
 * @brief it_end - Second iterator
 * @return Distance in characters between first and second ietators
 */
int64_t zm_iterator_get_distance(const ZM_ITERATOR  * it_start,
                                 const ZM_ITERATOR  * it_end);

/**
 * @brief Get current address
 * @brief it - Iterator object
 * @return Current iterator pointer
 */
const void * zm_iterator_get_addr(const ZM_ITERATOR  * it);

/**
 * @brief Get data between two iterators
 * @brief it_start - First iterator
 * @brief it_end - Second iterator
 * @brief len - Data length [out]
 * @return Pointer to the data
 */
const void * zm_iterator_get_data(const ZM_ITERATOR  * it_start,
                                  const ZM_ITERATOR  * it_end,
                                  uint64_t           * len);

/**
 * @brief Iterate next symbol
 * @brief it - Iterator object
 */
void zm_iterator_next(ZM_ITERATOR  * it);

/**
 * @brief Get current character
 * @brief it - Iterator object
 * @return Current charecter at iterator
 */
const unsigned char zm_iterator_ref(const ZM_ITERATOR  * it);

/**
 * @brief Get stream properties
 * @brief it - Iterator object
 * @brief bits - Max. stream bits
 * @brief line_len - Max. line length
 */
void zm_iterator_stream_props(ZM_ITERATOR   * it,
                              unsigned char * bits,
                              uint64_t      * line_len);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_ITERATOR_H__ */
/* End. */
