/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_MEM_H__
#define _ZM_MEM_H__ 1

/**
 * @file zm_mem.h
 * @brief Memory allocation/deallocation functions
 */

#include "zm_mem_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

/** Allocate memory function AND zero allocated memory block */
typedef void * (*zm_calloc_fn)(void    * opaque,
                               size_t    size);

/** Reallocate memory function */
typedef void * (*zm_realloc_fn)(void    * opaque,
                                void    * ptr,
                                size_t    size);

/** Free memory function */
typedef void (*zm_dealloc_fn)(void  * opaque,
                              void  * ptr);

/**
 * @struct _zm_allocator zm_mem.h <zm_mem.h>
 * @brief Memory allocator
 */
struct _zm_allocator
{
	/** Opaque data for allocator  */
	void           * opaque;
	/** Allocate memory            */
	zm_calloc_fn     allocate;
	/** Reallocate memory function */
	zm_realloc_fn    reallocate;
	/** Free memory                */
	zm_dealloc_fn    deallocate;
};

/**
 * @brief Allocate memory
 * @param alloc - Memory allocator
 * @param size - Block size
 * @return Memory block or NULL if any error occured
 */
void * zm_allocate(ZM_ALLOCATOR * alloc, size_t size);

/**
 * @brief Reallocate memory
 * @param alloc - Memory allocator
 * @param data - Old memory block
 * @param size - New block size
 * @return Memory block or NULL if any error occured
 */
void * zm_reallocate(ZM_ALLOCATOR * alloc, void * data, size_t size);

/**
 * @brief Deallocate memory
 * @param alloc - Memory allocator
 * @param data - Memory block to free
 */
void zm_deallocate(ZM_ALLOCATOR * alloc, void * data);

/**
 * @brief Duplicate memory block
 * @param alloc - Memory allocator
 * @param data - Memory block to free
 * @param size - Memory block size
 * @return Memory block or NULL if any error occured
 */
void * zm_duplicate(ZM_ALLOCATOR * alloc, const void * data, size_t size);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_MEM_H__ */
/* End. */
