/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_MIME_CONVERTER_H__
#define _ZM_MIME_CONVERTER_H__ 1

/**
  @file zm_mime_converter.h
  @brief Zero MIME MIME-encoded converter
*/

#include "zm_block_defs.h"
#include "zm_data_buffer_defs.h"
#include "zm_iconv_defs.h"
#include "zm_mem_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief RFC 3548 - The Base16, Base32, and Base64 Data Charsets for URLs, character map
 */
extern unsigned char zm_base64_rev[256];

/**
 * @brief Decode Base64 data block
 * @param src - source block
 * @param src_len - source block length
 * @param dst - destination block [out]
 * @param dst_len - destination block length [out]
 * @param max_dst_len - max. destination block length
 * @param base64_map character map
 * @return pointer to source block after processing
 */
const unsigned char  * zm_decode_base64_block(const unsigned char  * src,
                                              uint64_t               src_len,
                                              unsigned char        * dst,
                                              uint64_t             * dst_len,
                                              uint64_t               max_dst_len,
                                              const unsigned char  * base64_map);

/**
 * @brief Decode Base64 data
 * @param src_block - source block
 * @param src - source data
 * @param dst - destination data buffer [out]
 * @param base64_map character map
 * @return pointer to source block after processing
 */
const void * zm_decode_base64(const ZM_BLOCK       * src_block,
                              const void           * src,
                              const unsigned char  * base64_map,
                              ZM_DATA_BUFFER       * dst);

/**
 * @brief Decode quoted-printable data block
 * @param src - source block
 * @param src_len - source block length
 * @param dst - destination block [out]
 * @param dst_len - destination block length [out]
 * @param max_dst_len - max. destination block length
 * @return pointer to source block after processing
 */
const unsigned char  * zm_decode_quoted_printable_block(const unsigned char  * src,
                                                        uint64_t               src_len,
                                                        unsigned char        * dst,
                                                        uint64_t             * dst_len,
                                                        uint64_t               max_dst_len);

/**
 * @brief Decode quoted-printable data
 * @param src_block - source block
 * @param src - source data
 * @param dst - destination data buffer [out]
 * @return pointer to source block after processing
 */
const void * zm_decode_quoted_printable(const ZM_BLOCK  * src_block,
                                        const void      * src,
                                        ZM_DATA_BUFFER  * dst);

/**
 * @brief Decode uu-encoded data block
 * @param src - source block
 * @param src_len - source block length
 * @param dst - destination block [out]
 * @param dst_len - destination block length [out]
 * @param max_dst_len - max. destination block length
 * @return pointer to source block after processing
 */
const unsigned char  * zm_decode_uu_encoded_block(const unsigned char  * src,
                                                  uint64_t               src_len,
                                                  unsigned char        * dst,
                                                  uint64_t             * dst_len,
                                                  uint64_t               max_dst_len);


/**
 * @brief Decode uu-encoded data
 * @param src_block - source block
 * @param src - source data
 * @param dst - destination data buffer [out]
 * @return pointer to source block after processing
 */
const void * zm_decode_uu_encoded(const ZM_BLOCK   * src_block,
                                  const void       * src,
                                  ZM_DATA_BUFFER   * dst);

/**
 * @brief Decode q-encoded data block
 * @param src - source block
 * @param src_len - source block length
 * @param dst - destination block [out]
 * @param dst_len - destination block length [out]
 * @param max_dst_len - max. destination block length
 * @return pointer to source block after processing
 */
const unsigned char * zm_decode_qencoded_block(const unsigned char  * src,
                                               uint64_t               src_len,
                                               unsigned char        * dst,
                                               uint64_t             * dst_len,
                                               uint64_t               max_dst_len);

/**
 * @brief Decode q-encoded data
 * @param src_block - source block
 * @param src - source data
 * @param dst - destination data buffer [out]
 * @return pointer to source block after processing
 */
const void * zm_decode_qencoded(const ZM_BLOCK  * src_block,
                                const void      * src,
                                ZM_DATA_BUFFER  * dst);

/**
 * @brief Decode MIME-encoded part data
 * @param src - source block
 * @param src_len - source block length
 * @param dst - destination data buffer [out]
 * @param iconv - Encoding converter
 * @return pointer to source block after processing
 */
const unsigned char  * zm_decode_mime_encoded(const unsigned char  * src,
                                              uint64_t               src_len,
                                              ZM_DATA_BUFFER       * dst,
                                              ZM_ICONV             * iconv);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_MIME_CONVERTER_H__ */
/* End. */
