/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_RBTREE_H__
#define _ZM_RBTREE_H__ 1

/**
 * @file zm_rbtree.h
 * @brief Red-black tree implementation
 */

/*
 * Copyright (c) 2004-2005, 2007 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Adapted from the following code written by Emin Martinian:
 * http://web.mit.edu/~emin/www/source_code/red_black_tree/index.html
 * http://web.mit.edu/~emin/Desktop/ref_to_emin/www.old/source_code/red_black_tree/index.html
 *
 * Copyright (c) 2001 Emin Martinian
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that neither the name of Emin
 * Martinian nor the names of any contributors are be used to endorse or
 * promote products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "zm_mem_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct _zm_rb_tree    ZM_RB_TREE;
typedef struct _zm_rb_node    ZM_RB_NODE;
typedef struct _zm_rb_pair    ZM_RB_PAIR;

/** Compare elements. The comparison function must return an integer less than, equal to, or greater than zero if the first argument is considered to be respectively less than, equal to, or greater than the second. */
typedef int (*zm_rb_tree_cmp_fn)(void        * opaque,
                                 const void  * x,
                                 const void  * y);

/** Destroy element callback */
typedef void (*zm_rb_tree_destroy_fn)(void  * opaque,
                                      void  * key,
                                      void  * val);

/** Dump rb-tree function */
typedef int (*zm_rb_tree_dump_fn)(void        * opaque,
                                  const void  * key,
                                  const void  * val);

/**
 * @struct _rb_node_pair zm_rbtree.h <zm_rbtree.h>
 * @brief Red-black tree node key => value pair
 */
struct ZM_API _zm_rb_pair
{
	/** Key   */
	void  * key;
	/** Value */
	void  * val;
};

/**
 * @struct _zm_rb_node zm_rbtree.h <zm_rbtree.h>
 * @brief Red-black tree node
 */
struct ZM_API _zm_rb_node
{
	/** Red-black tree
	    node key => value
	                  pair */
	ZM_RB_PAIR    pair;
	/** Left child        */
	ZM_RB_NODE  * left;
	/** Right child       */
	ZM_RB_NODE  * right;
	/** Parent node       */
	ZM_RB_NODE  * parent;
	/** Flag              */
	uint32_t      red;
};

/**
 * @struct _rb_tree zm_rbtree.h <zm_rbtree.h>
 * @brief Red-black tree
 */
struct ZM_API _zm_rb_tree
{
	/** Root node                  */
	ZM_RB_NODE               root;
	/** NULL node                  */
	ZM_RB_NODE               nil;
	/* Tree size                   */
	uint32_t                 size;

	/** Opaque data for comparator
	                 and destructor
	                               */
	void                   * opaque;
	/** Comparator                 */
	zm_rb_tree_cmp_fn        cmp;
	/** Destroy node key and value */
	zm_rb_tree_destroy_fn    destroy;

	/** Opaque data allocator       */
	ZM_ALLOCATOR           * alloc;

	/* Tree flags                  */
	uint32_t                 flags;
};

/**
 * @brief Create tree
 * @param tree - If tree is a NULL pointer, the function allocates, initializes, and returns a new object. Otherwise, the object is initialized and the address of the object is returned.
 * @param opaque - Opaque data for comparator and destructor functions
 * @param cmp - Comparator
 * @param destroy - Destrustor for key and value
 * @param alloc - memory allocator
 * @return An initialized ZM_RB_TREE * handler. NULL if there was insufficient memory to allocate a new object.
 */
ZM_RB_TREE * zm_rb_create(ZM_RB_TREE             * tree,
                          void                   * opaque,
                          zm_rb_tree_cmp_fn        cmp,
                          zm_rb_tree_destroy_fn    destroy,
                          ZM_ALLOCATOR           * alloc);

/**
 * @brief Destroy tree, free memory and destruct all elements
 * @param tree - RB-tree
 */
void zm_rb_destroy(ZM_RB_TREE  * tree);

/**
 * @brief Insert data pointer into a redblack tree.
 * @param tree - RB-tree
 * @param key - Key
 * @param val - Value
 * @return RB-node, if node with following key exist, NULL - if inertion completed successfully
 */
ZM_RB_NODE * zm_rb_insert_node(ZM_RB_TREE  * tree,
                               void          * key,
                               void          * val);

/**
 * @brief Look for a node matching key in tree
 * @param tree - RB-tree
 * @param key - Key
 * @return A pointer to the node if found, else NULL
 */
ZM_RB_NODE * zm_rb_search_node(ZM_RB_TREE  * tree,
                               const void  * key);

/**
 * @brief Delete node from tree
 * @param tree - RB-tree
 * @param z - RB-tree node to delete
 */
void zm_rb_delete_node(ZM_RB_TREE  * tree,
                       ZM_RB_NODE  * z);

/**
 * @brief Insert data pointer into a redblack tree.
 * @param tree - RB-tree
 * @return
 */
int32_t zm_rb_insert(ZM_RB_TREE  * tree,
                     void        * key,
                     void        * val);

/**
 * @brief Look for a value matching key in tree. Returns a pointer to the node if found, else NULL.
 * @param tree - RB-tree
 * @param
 * @return
 */
ZM_RB_PAIR * zm_rb_search(ZM_RB_TREE  * tree,
                          const void  * key);

/**
 * @brief Delete node from tree
 * @param tree - RB-tree
 * @param
 * @return
 */
int32_t zm_rb_delete(ZM_RB_TREE  * tree,
                     const void  * key);

/**
 * @brief Clear tree, destruct all elements
 * @param tree - RB-tree
 */
void zm_rb_clear(ZM_RB_TREE  * tree);

/**
 * @brief Dump tree
 * @param tree - RB-tree
 * @param
 * @return
 */
int32_t zm_rb_dump(ZM_RB_TREE          * tree,
                   void                * opaque,
                   zm_rb_tree_dump_fn    dump);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_RBTREE_H__ */
/* End */
