/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#ifndef _ZM_STRUCT_H__
#define _ZM_STRUCT_H__ 1

/**
  @file zm_struct.h
  @brief Zero MIME data structures
*/

#include "zm_block.h"
#include "zm_mem_defs.h"
#include "zm_struct_defs.h"

#define ZM_FATAL           0x0001
#define ZM_FAILED          0x0002
#define ZM_FOUND           0x0004
#define ZM_BOUNDARY        0x0008
#define ZM_BOUNDARY_END    0x0010
#define ZM_LF              0x0020
#define ZM_END             0x0040

#define ZM_PART_NAME_LEN   256

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @struct _zm_key_val zm_struct.h <zm_struct.h>
 * @brief Key-value pair
 */
struct ZM_API _zm_key_val
{
	/** Key        */
	ZM_BLOCK      key;
	/** Value      */
	ZM_BLOCK      val;
	/** Next pair  */
	ZM_KEY_VAL  * next;
};

/**
 * @struct _zm_field_val zm_struct.h <zm_struct.h>
 * @brief Header value record
 */
struct ZM_API _zm_field_val
{
	/** Header value             */
	ZM_BLOCK      raw;

	/** Number of attributes     */
	uint32_t      pair_num;
	/** List of value attributes */
	ZM_KEY_VAL  * pairs;
};

/**
 * @struct _zm_field_rec zm_struct.h <zm_struct.h>
 * @brief Header field record
 */
struct ZM_API _zm_field_rec
{
	/** Header name              */
	ZM_BLOCK        name;
	/** Header value             */
	ZM_FIELD_VAL    value;

	/** Next header              */
	ZM_FIELD_REC  * next;
};

/**
 * @struct _zm_header zm_struct.h <zm_struct.h>
 * @brief Header field record
 */
struct ZM_API _zm_header
{
	/** Raw data of header section */
	ZM_BLOCK        raw;

	/** Number of fields           */
	uint32_t        fields_num;
	/** List of fields             */
	ZM_FIELD_REC  * fields;
};

/**
 * @struct _zm_part_data zm_struct.h <zm_struct.h>
 * @brief Part data
 */
struct ZM_API _zm_part_data
{
	/** Raw data        */
	ZM_BLOCK    raw;
	/** Number of lines */
	uint64_t    lines;
};

/* Base part (headers only) */
#define  ZM_BASE_PART_STRUCT \
	/** Part name                          */                            \
	uint32_t                             part_num;                       \
	/** Is text part                       */                            \
	uint32_t                             is_text:1;                      \
	/** Is basic part                      */                            \
	uint32_t                             is_basic:1;                     \
	/** Content type                       */                            \
	enum ZM_CONTENT_TYPE                 type;                           \
	/** Content type                       */                            \
	enum ZM_CONTENT_SUBTYPE              subtype;                        \
	                                                                     \
	/** Content transfer encoding          */                            \
	enum ZM_CONTENT_TRANSFER_ENCODING    content_transfer_encoding;      \
	/** Content disposition                */                            \
	enum ZM_CONTENT_DISPOSITION          content_disposition;            \
	                                                                     \
	/** Part type                          */                            \
	ZM_BLOCK                             raw_type;                       \
	/** Subpart type                       */                            \
	ZM_BLOCK                             raw_subtype;                    \
	/** Content type values                */                            \
	ZM_FIELD_VAL                       * content_type_val;               \
	                                                                     \
	/** Content disposotion                */                            \
	ZM_FIELD_VAL                         raw_content_disposition;        \
	/** Content disposotion filename       */                            \
	ZM_BLOCK                           * content_disposition_filename;   \
	/** Content transfer encoding          */                            \
	ZM_BLOCK                             raw_content_transfer_encoding;  \
	                                                                     \
	/** Content Language                   */                            \
	ZM_BLOCK                             raw_content_language;           \
	/** Content Location                   */                            \
	ZM_BLOCK                             raw_content_location;           \
	/** Content Id                         */                            \
	ZM_BLOCK                             raw_content_id;                 \
	/** Content Description                */                            \
	ZM_BLOCK                             raw_content_description;        \
	/** Content MD5                        */                            \
	ZM_BLOCK                             raw_content_md5;                \
	/** Message-Id                         */                            \
	ZM_BLOCK                             raw_message_id;                 \
	                                                                     \
	/** List of headers                    */                            \
	ZM_HEADER                            header;                         \
	/** Part data                          */                            \
	ZM_PART_DATA                         data

/**
 * @brief Basic part (image, audio, video, application, other)
 *
struct ZM_API _zm_basic_part
{
	;; / * No members * /
};
*/

/**
 * @struct _zm_text_part zm_struct.h <zm_struct.h>
 * @brief Text part
 */
struct ZM_API _zm_text_part
{
	/** Charset          */
	enum ZM_CHARSET    charset;
	/** Charset          */
	ZM_BLOCK           raw_charset;
};

/**
 * @struct _zm_multi_part zm_struct.h <zm_struct.h>
 * @brief Multipart part (mixed, alternative, related, report, parallel, digest, signed, encrypted, other)
 */
struct ZM_API _zm_multi_part
{
	/** Boundary            */
	ZM_BLOCK    boundary;
	/** Number of parts     */
	uint32_t    parts_num;
	/** List of parts       */
	ZM_PART   * parts;
};

/**
 * @struct _zm_message_part zm_struct.h <zm_struct.h>
 * @brief Message part (delivery-status, disposition-notification-to, external-body, partial, rfc822, other)
 */
struct ZM_API _zm_message_part
{
	/** Message part  */
	ZM_PART   * part;
};

/**
 * @struct _zm_part zm_struct.h <zm_struct.h>
 * @brief Generic part
 */
struct ZM_API _zm_part
{
	ZM_BASE_PART_STRUCT;
	union
	{
		/* ZM_BASIC_PART    basic; */
		/** Text part      */
		ZM_TEXT_PART     text;
		/** Multipart part */
		ZM_MULTI_PART    multi;
		/** Message part   */
		ZM_MESSAGE_PART  message;
	} u;

	/** Parent part   */
	ZM_PART       * parent;
	/** Next part     */
	ZM_PART       * next;
	/** Allocator     */
	ZM_ALLOCATOR  * alloc;
};

/**
 * @brief Get printable name of part content type
 * @param part - Generic part
 * @param msg - message data
 * @param type_len - part content type length [out]
 * @return Content type
 */
const char * zm_get_part_type(const ZM_PART  * part,
                              const void     * msg,
                              uint64_t       * type_len);

/**
 * @brief Get printable name of part content subtype
 * @param part - Generic part
 * @param msg - message data
 * @param subtype_len - part content subtype length [out]
 * @return Content subtype
 */
const char * zm_get_part_subtype(const ZM_PART  * part,
                                 const void     * msg,
                                 uint64_t       * subtype_len);

/**
 * @brief Get printable name of part content transfer encoding
 * @param part - Generic part
 * @param msg - message data
 * @param encoding_len - part content transfer encoding length [out]
 * @return Content transfer encoding
 */
const char * zm_get_part_content_transfer_encoding(const ZM_PART  * part,
                                                   const void     * msg,
                                                   uint64_t       * encoding_len);

/**
 * @brief Get printable name of part content disposition
 * @param part - Generic part
 * @param msg - message data
 * @param disposition_len - part content disposition length [out]
 * @return Content disposition
 */
const char * zm_get_part_content_disposition(const ZM_PART  * part,
                                             const void     * msg,
                                             uint64_t       * disposition_len);

/**
 * @brief Get printable name of character set
 * @param part - Text part
 * @param msg - message data
 * @param charset_len - character set length [out]
 * @return Character set
 */
const char * zm_get_text_part_charset(const ZM_TEXT_PART  * part,
                                      const void          * msg,
                                      uint64_t            * charset_len);

/**
 * @brief Get next record or NULL, if no next record available
 * @param key_val - Current key-value pair
 * @return Next key-value pair or NULL, if no next record available
 */
ZM_KEY_VAL * zm_key_val_next(const ZM_KEY_VAL  * key_val);

/**
 * @brief Get first key-value pair from header field value
 * @param field_val - Header field value
 * @return First key-value pair or NULL, if no next record available
 */
ZM_KEY_VAL * zm_field_val_first(const ZM_FIELD_VAL * field_val);

/**
 * @brief Get next header of part or NULL, if no next record available
 * @param header - Header value
 * @return Next header or NULL, if no next record available
 */
ZM_FIELD_REC * zm_part_header_next(const ZM_FIELD_REC  * header);

/**
 * @brief Get first header of multipart or NULL, if no headers available
 * @param mpart - current MIME part
 * @return First header or NULL, if no next record available
 */
ZM_FIELD_REC * zm_part_first_header(const ZM_PART  * part);

/**
 * @brief Get next part or NULL, if no next record available
 * @param part - MIME part
 * @return Next multipart subpart or NULL if no subpart/wrong part type
 */
ZM_PART * zm_next_part(const ZM_PART  * part);

/**
 * @brief Get first subpart of multipart or NULL, if no subparts  available/wrong part type
 * @param part - MIME MULTIPART part
 * @return Next multipart subpart or NULL if no subpart/wrong part type
 */
ZM_PART * zm_multipart_first_subpart(const ZM_PART  * part);

/**
 * @brief Get message subpart
 * @param part - MIME MESSAGE part
 * @return Message subpart or NULL if no subpart/wrong part type
 */
ZM_PART * zm_message_subpart(const ZM_PART  * part);

/**
 * @brief Free MIME part memrory and all subparts
 * @param part - MIME Part to free
 */
void zm_free_part(ZM_PART * part);

#ifdef __cplusplus
}
#endif

#endif /* _ZM_STRUCT_H__ */
/* End. */
