/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_address_parser.h"
#include "zm_data_buffer.h"
#include "zm_email_address.h"
#include "zm_iconv.h"
#include "zm_mem.h"
#include "zm_mime_converter.h"

#include <stdio.h>

#define DEBUG 1

#ifndef DEBUG_FPRINTF
    #ifdef DEBUG
        #define DEBUG_FPRINTF fprintf
    #else
        #define DEBUG_FPRINTF zm_do_nothing
        void zm_do_nothing(FILE * stream, const char * format, ...) { ;; }
    #endif  /* DEBUG */
#endif  /* DEBUG_FPRINTF */

/*
 * fws = wsp | crlf
 */
static int32_t zm_address_parser_is_fws(const unsigned char ** src,
                                        uint64_t             * src_len)
{
	unsigned char ch;

	for(;;)
	{
		if (*src_len == 0) { return ZM_ADDRESS_EOL; }

		ch = **src;
		if (ch != ' ' && ch != '\t' && ch != '\r' && ch != '\n') { break; }

		++*src; --*src_len;
	}

return ZM_ADDRESS_OK;
}

/*
 * cfws = (1 * ([fws] comment) [fws]) | fws
 */
static int32_t zm_address_parser_is_cfws(const unsigned char ** src,
                                         uint64_t             * src_len)
{
	const unsigned char * start = *src;
	for (;;)
	{
		unsigned char ch;

		if (*src_len == 0) { return ZM_ADDRESS_EOL; }

		ch = **src;
		if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
		{
			++*src; --*src_len;
		}
		else if (ch == '(')
		{
			int32_t level = 1;
			++*src; --*src_len;

			for(;;)
			{
				if (*src_len == 0) { return ZM_ADDRESS_ERROR; }

				ch = **src;
				++*src; --*src_len;

				if (ch == '(')
				{
					++level;
				}
				else if (ch == ')')
				{
					--level;
					if (level == 0) { break; }
				}
				else if (ch == '\\')
				{
					if (*src_len == 0) { return ZM_ADDRESS_ERROR; }
					++*src; --*src_len;
				}
			}

		}
		else { break; }
	}
return (start != *src) ? ZM_ADDRESS_OK : ZM_ADDRESS_NO_CFWS;
}

/*
 * atext = ALPHA  | DIGIT |
 *            "!" | "#"   |
 *            "$" | "%"   |
 *            "&" | "'"   |
 *            "*" | "+"   |
 *            "-" | "/"   |
 *            "=" | "?"   |
 *            "^" | "_"   |
 *            "`" | "{"   |
 *            "|" | "}"   |
 *            "~"
 */
static int32_t zm_address_parser_is_atext(const unsigned char ** src,
                                          uint64_t             * src_len)
{
	unsigned char ch;

	if (*src_len == 0) { return ZM_ADDRESS_EOL; }

	ch = **src;
	if ((ch >= 'A' && ch <= 'Z') ||
	    (ch >= 'a' && ch <= 'z') ||
	    (ch >= '0' && ch <= '9') ||

	     ch == '!' || ch == '#'  ||
	     ch == '$' || ch == '%'  ||
	     ch == '&' || ch == '\'' ||
	     ch == '*' || ch == '+'  ||
	     ch == '-' || ch == '/'  ||
	     ch == '=' || ch == '?'  ||
	     ch == '^' || ch == '_'  ||
	     ch == '`' || ch == '{'  ||
	     ch == '|' || ch == '}'  ||
	     ch == '~' || ch == '{'  ||
	     ch == '`')
	{
		++*src; --*src_len;
		return ZM_ADDRESS_OK;
	}
return ZM_ADDRESS_ERROR;
}

/*
 * atom = atext *(atext)
 */
static int32_t zm_address_parser_is_atom(const unsigned char ** src,
                                         uint64_t             * src_len,
                                         ZM_DATA_BUFFER       * atom)
{
	const unsigned char  * start = *src;

	int32_t rc = zm_address_parser_is_atext(src, src_len);
	if (rc != ZM_ADDRESS_OK) { return rc; }

	for(;;)
	{
		rc = zm_address_parser_is_atext(src, src_len);
		if (rc != ZM_ADDRESS_OK) { break; }
	}

	zm_data_buffer_append_data(atom, start, *src - start);

return ZM_ADDRESS_OK;
}

/*
 * quoted-string = '"' *(any - '"') '"'
 */
static int32_t zm_address_parser_is_quoted_string(const unsigned char ** src,
                                                  uint64_t             * src_len,
                                                  ZM_DATA_BUFFER       * quoted_string)
{
	const unsigned char  * start;

	if (*src_len == 0) { return ZM_ADDRESS_EOL; }

	if (**src != '"') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

	start = *src;
	for(;;)
	{
		unsigned char ch;

		if (*src_len == 0) { return ZM_ADDRESS_ERROR; }

		ch = **src;
		if (ch == '"') { break; }

		if (ch == '\\')
		{
			if (*src_len == 0) { return ZM_ADDRESS_ERROR; }

			zm_data_buffer_append_data(quoted_string, start, *src - start);

			++*src; --*src_len;
			start = *src;
		}

		++*src; --*src_len;
	}

	zm_data_buffer_append_data(quoted_string, start, *src - start);
	++*src; --*src_len;

return ZM_ADDRESS_OK;
}

/*
 * word = atom | quoted-string
 */
static int32_t zm_address_parser_is_word(const unsigned char ** src,
                                         uint64_t             * src_len,
                                         ZM_DATA_BUFFER       * word)
{
	int32_t rc = zm_address_parser_is_atom(src, src_len, word);
	if (rc == ZM_ADDRESS_OK || rc == ZM_ADDRESS_EOL) { return rc; }

return zm_address_parser_is_quoted_string(src, src_len, word);
}

/*
 * phrase = 1*word | word *(word | "." | cfws)
 */
static int32_t zm_address_parser_is_phrase(const unsigned char ** src,
                                           uint64_t             * src_len,
                                           ZM_DATA_BUFFER       * phrase)
{
	int32_t flag = 0;
	for(;;)
	{
		int32_t rc = zm_address_parser_is_word(src, src_len, phrase);
		if (rc != ZM_ADDRESS_OK)
		{
			if (flag == 1)
			{
				zm_data_buffer_set_size(phrase, zm_data_buffer_get_size(phrase) - 1);
				return ZM_ADDRESS_OK;
			}
			return rc;
		}
		flag = 1;

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS)
		{
			zm_data_buffer_append_char(phrase, ' ');
		}

		if (**src == '.')
		{
			zm_data_buffer_append_char(phrase, '.');
			++*src; --*src_len;
		}

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS)
		{
			zm_data_buffer_append_char(phrase, ' ');
		}

	}

/* Make compiler happy */
return ZM_ADDRESS_OK;
}

/*
 * display-name = phrase
 */
static int32_t zm_address_parser_is_display_name(const unsigned char ** src,
                                                 uint64_t             * src_len,
                                                 ZM_DATA_BUFFER       * display_name)
{
	;;
return zm_address_parser_is_phrase(src, src_len, display_name);
}

/*
 * local-part = dot-atom | quoted-string | obs-local-part
 */
static int32_t zm_address_parser_is_local_part(const unsigned char ** src,
                                               uint64_t             * src_len,
                                               ZM_DATA_BUFFER       * local_part)
{
	int32_t flag = 0;
	for(;;)
	{
		int32_t rc = zm_address_parser_is_atom(src, src_len, local_part);
		if (rc != ZM_ADDRESS_OK)
		{
			rc = zm_address_parser_is_word(src, src_len, local_part);
			if (rc != ZM_ADDRESS_OK)
			{
				rc = zm_address_parser_is_quoted_string(src, src_len, local_part);
				if (rc != ZM_ADDRESS_OK)
				{
					return flag == 1 ? ZM_ADDRESS_OK : rc;
				}
			}
		}
		flag = 1;

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS) { return rc; }

		if (**src == '.')
		{
			zm_data_buffer_append_char(local_part, '.');
			++*src; --*src_len;
		}

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS) { return rc; }
	}

/* Make compiler happy */
return ZM_ADDRESS_OK;
}

/*
 * dtext = %d33-90 | %d94-127
 */
static int32_t zm_address_parser_is_dtext(const unsigned char ** src,
                                          uint64_t             * src_len)
{
	unsigned char ch;

	if (*src_len == 0) { return ZM_ADDRESS_EOL; }

	ch = **src;

	if ((ch >= 33 && ch <= 90) || ch >= 94)
	{
		++*src; --*src_len;
		return ZM_ADDRESS_OK;
	}

return ZM_ADDRESS_ERROR;
}

/*
 * domain-literal  = "[" *([fws] dtext) [fws] "]"
 */
static int32_t zm_address_parser_is_domain_literal(const unsigned char ** src,
                                                   uint64_t             * src_len,
                                                   ZM_DATA_BUFFER       * domain_literal)
{
	int32_t rc = 0;
	const unsigned char  * start;
	const unsigned char  * end;

	if (*src_len == 0) { return ZM_ADDRESS_EOL; }

	if (**src != '[') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

	rc = zm_address_parser_is_fws(src, src_len);
	if (rc != ZM_ADDRESS_OK) { return ZM_ADDRESS_ERROR; }

	start = *src;

	for(;;)
	{
		rc = zm_address_parser_is_dtext(src, src_len);
		if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_ERROR; }
		if (rc != ZM_ADDRESS_OK) { break; }
	}
	end = *src;

	rc = zm_address_parser_is_fws(src, src_len);
	if (rc != ZM_ADDRESS_OK) { return ZM_ADDRESS_ERROR; }

	if (**src != ']') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

	zm_data_buffer_append_char(domain_literal, '[');
	zm_data_buffer_append_data(domain_literal, start, end - start);
	zm_data_buffer_append_char(domain_literal, ']');

return ZM_ADDRESS_OK;
}

/*
 * domain = dot-atom | domain-literal | obs-domain
 */
static int32_t zm_address_parser_is_domain(const unsigned char ** src,
                                           uint64_t             * src_len,
                                           ZM_DATA_BUFFER       * domain)
{
	int32_t flag = 0;

	int32_t rc = zm_address_parser_is_domain_literal(src, src_len, domain);
	if (rc == ZM_ADDRESS_OK) { return ZM_ADDRESS_OK; }

	for(;;)
	{
		int32_t rc = zm_address_parser_is_atom(src, src_len, domain);
		if (rc != ZM_ADDRESS_OK)
		{
			return flag == 1 ? ZM_ADDRESS_OK : rc;
		}
		flag = 1;

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS) { return rc; }

		if (**src == '.')
		{
			zm_data_buffer_append_char(domain, '.');
			++*src; --*src_len;
		}

		/* break after cfws */
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc != ZM_ADDRESS_NO_CFWS) { return rc; }
	}

/* Make compiler happy */
return ZM_ADDRESS_OK;
}

/*
 * addr-spec = local-part "@" domain
 */
static int32_t zm_address_parser_is_addr_spec(const unsigned char ** src,
                                              uint64_t             * src_len,
                                              ZM_EMAIL_ADDRESS     * email_address)
{
	int32_t rc = zm_address_parser_is_local_part(src, src_len, &(email_address -> username));
	if (rc != ZM_ADDRESS_OK) { return rc; }

	if (**src != '@') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

return zm_address_parser_is_domain(src, src_len, &(email_address -> domain));
}

/*
 * obs-domain-list = *(cfws | ",") "@" domain *("," [cfws] ["@" domain])
 * obs-route = obs-domain-list ":"
 */
static int32_t zm_address_parser_is_route(const unsigned char ** src,
                                          uint64_t             * src_len,
                                          ZM_EMAIL_ADDRESS     * email_address)
{
	int32_t rc = 0;
	int32_t flag = 0;

	for(;;)
	{
		for(;;)
		{
			rc = zm_address_parser_is_cfws(src, src_len);
			if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

			if (**src != ',') { break; }
			++*src; --*src_len;
		}

		if (*src_len == 0) { return ZM_ADDRESS_EOL; }

		if (**src == ':') { break; }

		if (**src != '@') { return ZM_ADDRESS_ERROR; }
		++*src; --*src_len;

		if (flag == 1) { zm_data_buffer_append_char(&(email_address -> route), ' '); }

		rc = zm_address_parser_is_domain(src, src_len, &(email_address -> route));
		if (rc != ZM_ADDRESS_OK) { return ZM_ADDRESS_ERROR; }

		if (flag == 0) { flag = 1; }
	}

	++*src; --*src_len;
return ZM_ADDRESS_OK;
}

/*
 * angle-addr = "<" [route] addr-spec ">"
 */
static int32_t zm_address_parser_is_angle_addr(const unsigned char ** src,
                                               uint64_t             * src_len,
                                               ZM_EMAIL_ADDRESS     * email_address)
{
	int32_t rc = 0;
	/* Store position */
	const unsigned char  * start;
	uint64_t               start_len;

	if (*src_len == 0) { return ZM_ADDRESS_EOL; }

	if (**src != '<') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

	rc = zm_address_parser_is_cfws(src, src_len);
	if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

	start     = *src;
	start_len = *src_len;

	rc = zm_address_parser_is_route(src, src_len, email_address);
	if (rc != ZM_ADDRESS_OK)
	{
		*src     = start;
		*src_len = start_len;
	}
	else
	{
		rc = zm_address_parser_is_cfws(src, src_len);
		if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }
	}

	rc = zm_address_parser_is_addr_spec(src, src_len, email_address);
	if (rc != ZM_ADDRESS_OK) { return ZM_ADDRESS_ERROR; }

	zm_address_parser_is_cfws(src, src_len);
	if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

	if (**src != '>') { return ZM_ADDRESS_ERROR; }
	++*src; --*src_len;

return ZM_ADDRESS_OK;
}

/*
 * name-addr = [display-name] angle-addr
 */
static int32_t zm_address_parser_is_name_addr(const unsigned char ** src,
                                              uint64_t             * src_len,
                                              ZM_EMAIL_ADDRESS     * email_address)
{
	/* Store position */
	const unsigned char  * start     = *src;
	uint64_t               start_len = *src_len;

	int32_t rc = zm_address_parser_is_display_name(src, src_len, &(email_address -> name));

	if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

	if (rc != ZM_ADDRESS_OK)
	{
		*src     = start;
		*src_len = start_len;
		return rc;
	}

	rc = zm_address_parser_is_cfws(src, src_len);
	if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

return zm_address_parser_is_angle_addr(src, src_len, email_address);
}

/*
 * mailbox = name-addr | addr-spec
 */
static int32_t zm_address_parser_is_mailbox(const unsigned char ** src,
                                            uint64_t             * src_len,
                                            ZM_EMAIL_ADDRESS     * email_address)
{
	int32_t rc;
	/* Store position */
	const unsigned char  * start;
	uint64_t               start_len;

	zm_address_parser_is_cfws(src, src_len);

	start     = *src;
	start_len = *src_len;

	rc = zm_address_parser_is_name_addr(src, src_len, email_address);
	if (rc == ZM_ADDRESS_OK) { return rc; }

	*src     = start;
	*src_len = start_len;
	zm_email_address_reset(email_address);

return zm_address_parser_is_addr_spec(src, src_len, email_address);
}

/*
 * Create email address parser
 */
ZM_ADDRESS_PARSER * zm_address_parser_create(ZM_ADDRESS_PARSER  * address_parser,
                                             ZM_ICONV           * iconv,
                                             ZM_DATA_BUFFER     * buffer)
{
	uint32_t        flags = 0x00;
	ZM_ALLOCATOR  * alloc = zm_data_buffer_get_allocator(buffer);

	if (address_parser == NULL)
	{
		address_parser = (ZM_ADDRESS_PARSER *)zm_allocate(alloc, sizeof(ZM_ADDRESS_PARSER));
		flags = 0x01;
	}

	if (address_parser == NULL) { return NULL; }

	if (zm_data_buffer_clone(&(address_parser -> comment), buffer) != NULL) { flags |= 0x10; }

	address_parser -> iconv  = iconv;
	address_parser -> alloc  = alloc;
	address_parser -> flags  = flags;

	if ((flags & 0x10) != 0x10)
	{
		zm_address_parser_destroy(address_parser);
		return NULL;
	}

return address_parser;
}

/*
 * Destroy address parser
 */
void zm_address_parser_destroy(ZM_ADDRESS_PARSER  * address_parser)
{
	ZM_ALLOCATOR  * alloc = address_parser -> alloc;

	if (address_parser -> flags & 0x10)
	{
		zm_data_buffer_destroy(&(address_parser -> comment));
	}

	if (address_parser -> flags & 0x01)
	{
		zm_deallocate(alloc, address_parser);
	}
}

/*
 * address-list  = (address *("," address)) | obs-addr-list
 * obs-addr-list = *([cfws] ",") address *("," [address / cfws])
 */
int32_t zm_address_parser_parse_address_list(ZM_ADDRESS_PARSER    * address_parser,
                                             const unsigned char ** src,
                                             uint64_t             * src_len,
                                             ZM_EMAIL_ADDRESS    ** email_address)
{
	int32_t rc;
	ZM_EMAIL_ADDRESS  * curr = *email_address;

	for(;;)
	{
		ZM_EMAIL_ADDRESS  * tmp = zm_email_address_create(NULL, &(address_parser -> comment));
		if (tmp == NULL) { return ZM_ADDRESS_ERROR; }

		rc = zm_address_parser_is_mailbox(src, src_len, tmp);
		if (rc != ZM_ADDRESS_OK && rc != ZM_ADDRESS_EOL)
		{
			zm_email_address_destroy(tmp);
			return rc;
		}

		if (curr == NULL)
		{
			curr = *email_address = tmp;
		}
		else
		{
			curr -> next = tmp;
			curr = tmp;
		}

		if (rc == ZM_ADDRESS_EOL) { return rc; }

		for(;;)
		{
			rc = zm_address_parser_is_cfws(src, src_len);
			if (rc == ZM_ADDRESS_EOL) { return ZM_ADDRESS_EOL; }

			if (**src != ',') { break; }
			++*src; --*src_len;
		}
	}

return 0;
}
/*
dot-atom        =   [cfws] dot-atom-text [cfws]
specials        =   "(" / ")" /        ; Special characters that do
                    "<" / ">" /        ;  not appear in atext
                    "[" / "]" /
                    ":" / ";" /
                    "@" / "\" /
                    "," / "." /
                    DQUOTE
qtext           =   %d33 /             ; Printable US-ASCII
                    %d35-91 /          ;  characters not including
                    %d93-126 /         ;  "\" or the quote character
                    obs-qtext

quoted-string   =  DQUOTE *([fws] qcontent) [fws] DQUOTE

word            =   atom | quoted-string

phrase          =   1*word | obs-phrase

display-name    =   phrase

obs-local-part  =   word *("." word)

local-part      =   dot-atom | quoted-string | obs-local-part

domain-literal  =   [cfws] "[" *([fws] dtext) [fws] "]" [cfws]

obs-domain      =   atom *("." atom)

domain          =   dot-atom | domain-literal | obs-domain

addr-spec       =   local-part "@" domain

obs-domain-list =   *(cfws | ",") "@" domain *("," [cfws] ["@" domain])

obs-route       =   obs-domain-list ":"

obs-angle-addr  =   [cfws] "<" obs-route addr-spec ">" [cfws]

angle-addr      =   [cfws] "<" addr-spec ">" [cfws] | obs-angle-addr

name-addr       =   [display-name] angle-addr
mailbox         =   name-addr | addr-spec

address         =   mailbox | group






mailbox-list    =   (mailbox *("," mailbox)) | obs-mbox-list
obs-mbox-list   =   *([cfws] ",") mailbox *("," [mailbox / cfws])

address-list    =   (address *("," address)) | obs-addr-list
obs-addr-list   =   *([cfws] ",") address *("," [address / cfws])

obs-group-list  =   1*([cfws] ",") [cfws]
group-list      =   mailbox-list | cfws | obs-group-list

group           =   display-name ":" [group-list] ";" [cfws]


*/

/* End */

