/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_mem.h"
#include "zm_aliases.h"

#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

#define ALIASES_LINE_SIZE    512
#define ZM_ALIAS_BUF_LEN     128

typedef struct _zm_aliases_rec    ZM_ALIASES_REC;

struct _zm_aliases_rec
{
	/** Alias     */
	char    alias[32];
	/** Name      */
	char    name[32];
};

struct _zm_aliases
{
	/** Abbreviation       */
	ZM_ALIASES_REC  * aliases;
	/** Timezone offset    */
	uint32_t          aliases_count;
	/** Allocator          */
	ZM_ALLOCATOR    * alloc;
};

typedef struct _zm_aliases_tmp    ZM_ALIASES_TMP;
struct _zm_aliases_tmp
{
	/** Alias              */
	ZM_ALIASES_REC    rec;
	/** Next alias or NULL */
	ZM_ALIASES_TMP  * next;
};

static void zm_close_aliases(FILE           * f,
                            ZM_ALLOCATOR   * alloc,
                            ZM_ALIASES_TMP  * aliases,
                            int32_t          err_no)
{
	fclose(f);
	while (aliases != NULL)
	{
		ZM_ALIASES_TMP * tmp = aliases -> next;
		zm_deallocate(alloc, aliases);
		aliases = tmp;
	}
	errno = err_no;
}

/*
 * Sorting functor
 */
static int zm_aliases_cmp(const void * x, const void * y)
{
	ZM_ALIASES_REC  * rec_x = (ZM_ALIASES_REC *)x;
	ZM_ALIASES_REC  * rec_y = (ZM_ALIASES_REC *)y;

return strcasecmp(rec_x -> alias, rec_y -> alias);
}

/*
 * Load custom timezone aliasesiatures
 */
int32_t zm_load_aliases(const char     * file,
                        ZM_ALLOCATOR   * alloc,
                        ZM_ALIASES    ** aliases)
{
	ZM_ALIASES_REC  * aliases_rec     = NULL;
	ZM_ALIASES_TMP  * aliases_tmp_st  = NULL;
	ZM_ALIASES_TMP  * aliases_tmp_cur = NULL;

	int32_t count = 0;

	FILE * f = fopen(file, "r");
	if (f == NULL) { return -1; }

	for(;;)
	{
		ZM_ALIASES_TMP  * tmp = NULL;
		char     buf[ALIASES_LINE_SIZE + 1];
		char     alias[ALIASES_LINE_SIZE];
		char     name[ALIASES_LINE_SIZE];

		uint32_t alias_len = 0;
		uint32_t name_len  = 0;

		if (fgets(buf, ALIASES_LINE_SIZE, f) == NULL) { break; }

		if (buf[0] == '#' || buf[0] == '\r' || buf[0] == '\n' || buf[0] == '\0') { continue; }

		/* ASCII    US-ASCII */
		if (sscanf(buf, "%s %s", alias, name) != 2)
		{
			zm_close_aliases(f, alloc, aliases_tmp_st, EINVAL);
			return -1;
		}

		alias_len = strlen(alias);
		if (alias_len > 31)
		{
			zm_close_aliases(f, alloc, aliases_tmp_st, EINVAL);
			return -1;
		}

		name_len = strlen(name);
		if (name_len > 31)
		{
			zm_close_aliases(f, alloc, aliases_tmp_st, EINVAL);
			return -1;
		}

		tmp = zm_allocate(alloc, sizeof(ZM_ALIASES_TMP));
		if (tmp == NULL)
		{
			zm_close_aliases(f, alloc, aliases_tmp_st, ENOMEM);
			return -1;
		}

		if (aliases_tmp_st == NULL)
		{
			aliases_tmp_st = aliases_tmp_cur = tmp;
		}
		else
		{
			aliases_tmp_cur -> next = tmp;
			aliases_tmp_cur = tmp;
		}

		++count;

		strncpy(aliases_tmp_cur -> rec.alias, alias, alias_len + 1);
		strncpy(aliases_tmp_cur -> rec.name,  name,  name_len + 1);
	}

	*aliases = zm_allocate(alloc, sizeof(ZM_ALIASES));
	if (*aliases == NULL)
	{
		zm_close_aliases(f, alloc, aliases_tmp_st, ENOMEM);
		return -1;
	}

	(*aliases) -> alloc = alloc;
	(*aliases) -> aliases_count = count;
	(*aliases) -> aliases       = aliases_rec = zm_allocate(alloc, count * sizeof(ZM_ALIASES_REC));
	if ((*aliases) -> aliases == NULL)
	{
		zm_close_aliases(f, alloc, aliases_tmp_st, ENOMEM);
		zm_deallocate(alloc, (*aliases));
		return -1;
	}

	while (aliases_tmp_st != NULL)
	{
		ZM_ALIASES_TMP * tmp = aliases_tmp_st -> next;

		memcpy(aliases_rec, &(aliases_tmp_st -> rec), sizeof(ZM_ALIASES_REC));

		zm_deallocate(alloc, aliases_tmp_st);
		aliases_tmp_st = tmp;
		++aliases_rec;
	}

	qsort((*aliases) -> aliases, count, sizeof(ZM_ALIASES_REC), zm_aliases_cmp);

	fclose(f);
return 0;
}

/*
 * Print custom timezone aliasesiatures
 */
void zm_print_aliases(ZM_ALIASES  * aliases)
{
	uint32_t pos = 0;

	ZM_ALIASES_REC * rec = aliases -> aliases;
	for(pos = 0; pos < aliases -> aliases_count; ++pos)
	{
		fprintf(stdout, "{ %19s, %12s }\n", rec -> alias, rec -> name);
		++rec;
	}
}

/*
 * Print custom timezone aliasesiatures
 */
void zm_free_aliases(ZM_ALIASES  * aliases)
{
	ZM_ALLOCATOR  * alloc = aliases -> alloc;
	zm_deallocate(alloc, aliases -> aliases);
	zm_deallocate(alloc, aliases);
}

/*
 * Get timezone offset
 */
static ZM_ALIASES_REC * zm_get_aliases_rec(const char      * alias,
                                           ZM_ALIASES_REC  * aliases_rec,
                                           uint32_t          aliases_rec_count)
{
	int32_t l_pos = 0;
	int32_t r_pos = aliases_rec_count;
	int32_t pos = 0;
	int32_t rc = 0;
	for(;;)
	{
		pos = (l_pos + r_pos) / 2;
		rc = strcasecmp(aliases_rec[pos].alias, alias);

		if (pos == l_pos || pos == r_pos) { return NULL; }

		if      (rc == 0) { break; }
		else if (rc <  0) { l_pos = pos; }
		else              { r_pos = pos; }
	}

return &aliases_rec[pos];
}

/*
 * Parse aliases string
 */
int32_t zm_parse_alias(ZM_ALIASES      * aliases,
                       const char      * alias,
                       const uint32_t    alias_len,
                       const char     ** name)
{
	ZM_ALIASES_REC  * aliases_rec = NULL;

	char alias_buf[ZM_ALIAS_BUF_LEN + 1];

	if (alias_len >= ZM_ALIAS_BUF_LEN) { return -1; }

	memcpy(alias_buf, alias, alias_len);
	alias_buf[alias_len] = 0;

	aliases_rec = zm_get_aliases_rec(alias_buf, aliases -> aliases, aliases -> aliases_count);
	if (aliases_rec == NULL) { return -1; }

	*name = aliases_rec -> name;

return 0;
}
/* End. */
