/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_charset_detector.h"

#include "zm_cp1251table.h"
#include "zm_koi8table.h"

#ifdef _DEBUG_CHARSET_DETECTOR
    #include <stdio.h>
#endif /* _DEBUG_CHARSET_DETECTOR */

/*
 * Convert UTF-8 character to wide
 */
static char zm_utf8_to_wide(const unsigned char ** msg,
                            uint64_t             * msg_len,
                            uint32_t             * ucs)
{
	const unsigned char  * msg_it   = *msg;
	unsigned char          prefix   = 0;
	int32_t                char_len = 0;
	int64_t                boundary = 0;
	uint64_t               pos = 0;

	/* No data */
	if (*msg == NULL || *msg_len == 0) { return 0; }

	/* Determine size of UTF8 character & check "shortest form" of char */
	prefix = *msg_it;

	/* 10000000b  & 00000000b */
	if ((prefix & 0x80) == 0)         { prefix = 0x7F; char_len = 1; boundary = 0x00000000; }
	/* 11100000b & 11000000b */
	else if ((prefix & 0xE0) == 0xC0) { prefix = 0x1F; char_len = 2; boundary = 0x00000080; }
	/* 11110000b & 11100000b */
	else if ((prefix & 0xF0) == 0xE0) { prefix = 0x0F; char_len = 3; boundary = 0x00000800; }
	/* 11111000b & 11110000b */
	else if ((prefix & 0xF8) == 0xF0) { prefix = 0x07; char_len = 4; boundary = 0x00010000; }
	/* 11111100b & 11111000b */
	else if ((prefix & 0xFC) == 0xF8) { prefix = 0x03; char_len = 5; boundary = 0x00200000; }
	/* 11111110b & 11111100b */
	else if ((prefix & 0xFE) == 0xFC) { prefix = 0x01; char_len = 6; boundary = 0x04000000; }
	/* Invalid UTF8 sequence */
	else { return -1; }

	/* Invalid UTF8 sequence */
	if (*msg_len < char_len) { return -1; }

	/* Special case for first character */
	prefix = (*msg_it) & prefix;
	*ucs = prefix;

	++msg_it;
	for (pos = 1; pos < char_len; ++pos)
	{
		/* Incorect characters in middle of string */
		if (((*msg_it) & 0xC0) != 0x80) { return -1; }

		/* Join value from 6-bit chunks */
		*ucs <<= 6;
		*ucs |= (*msg_it) & 0x3F;
		++msg_it;
	}

	/* Check boundary */
	if (*ucs < boundary)
	{
		/* Modified UTF-8, special case */
		if (!(*ucs == 0 && boundary == 0x80))
		{
			/* Not a "shortest form" of char */
			return -1;
		}
	}

	*msg      = msg_it;
	*msg_len -= char_len;

return char_len;
}

/*
 * Get length of string
 */
static int64_t zm_utf8_length(const unsigned char  * msg,
                              const uint64_t         msg_len,
                              int32_t              * seven_bit_only)
{
	const unsigned char  * msg_tmp      = msg;
	uint64_t               msg_len_tmp  = msg_len;
	int64_t                msg_utf8_len = 0;

	for (;;)
	{
		uint32_t ucs = 0;
		char char_len = zm_utf8_to_wide(&msg_tmp, &msg_len_tmp, &ucs);

		if      (char_len == -1) { return -1; }
		else if (char_len >   1) { *seven_bit_only = 0; }

		msg_utf8_len += 1;

		if (msg_len_tmp == 0) { return msg_utf8_len; }
	}

/* Make compiler happy */
return 0;
}


/*
 * Detect charset, one of utf-8, koi8, cp1251
 */
enum ZM_CHARSET zm_detect_charset(const unsigned char  * msg,
                                  const uint64_t         msg_len)
{
	const unsigned char  * msg_end = msg + msg_len;
	unsigned char          ch_prev = 0;

	uint64_t us_ascii_weight = 0;
	uint64_t koi8_weight     = 0;
	uint64_t cp1251_weight   = 0;
	int32_t  seven_bit_only  = 1;

	/* Check UTF-8 */
	if (zm_utf8_length(msg, msg_len, &seven_bit_only) > 0)
	{
		if (seven_bit_only == 1) { return US_ASCII; }
		return UTF_8;
	}

	while (msg != msg_end)
	{
		unsigned char ch = *msg;
		if (ch <= 128) { us_ascii_weight += 1; }
		else
		{
			ch -= 128;
			ch |= 0x20;

			/* Check KOI-8 */
			koi8_weight   += zm_stat_koi8[ch_prev][ch];
			/* Check CP1251 */
			cp1251_weight += zm_stat_cp1251[ch_prev][ch];
		}

		ch_prev = ch;
		++msg;
	}

#ifdef _DEBUG_CHARSET_DETECTOR
	fprintf(stderr, "koi8_weight %llu, cp1251_weight %llu\n", (long long unsigned)koi8_weight, (long long unsigned)cp1251_weight);
#endif /* _DEBUG_CHARSET_DETECTOR */

	if (koi8_weight > 0 && cp1251_weight > 0)
	{
		if (koi8_weight > cp1251_weight) { return KOI_8R; }
		return WINDOWS_1251;
	}

	if (us_ascii_weight != 0) { return US_ASCII; }

return OTHER_CHARSET;
}
/* End */
