/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_data_buffer.h"
#include "zm_data_buffer_base.h"
#include "zm_data_buffer_impl.h"
#include "zm_mem.h"

#include <string.h>

/*
 * Create data buffer using existing object type and allocator
 */
ZM_DATA_BUFFER * zm_data_buffer_clone(ZM_DATA_BUFFER  * data_buffer,
                                      ZM_DATA_BUFFER  * data_buffer_src)
{
	uint32_t        flags  = 0x00;
	ZM_ALLOCATOR  * alloc  = zm_data_buffer_get_allocator(data_buffer_src);

	if (data_buffer == NULL)
	{
		data_buffer = (ZM_DATA_BUFFER *)zm_allocate(alloc, sizeof(ZM_DATA_BUFFER));
		flags |= 0x01;
	}

	if (data_buffer == NULL) { return NULL; }

	data_buffer -> vt     = data_buffer_src -> vt;
	data_buffer -> flags  = flags;
	data_buffer -> opaque = data_buffer_src -> vt -> create_fn(alloc);
	if (data_buffer -> opaque == NULL)
	{
		zm_deallocate(alloc, data_buffer);
		return NULL;
	}

return data_buffer;
}

/*
 * Destroy data_buffer
 */
void zm_data_buffer_destroy(ZM_DATA_BUFFER  * data_buffer)
{
	ZM_ALLOCATOR  * alloc  = zm_data_buffer_get_allocator(data_buffer);

	data_buffer -> vt -> destroy_fn(data_buffer -> opaque);

	if (data_buffer -> flags & 0x01)
	{
		zm_deallocate(alloc, data_buffer);
	}
}

/*
 * Get allocator
 */
ZM_ALLOCATOR * zm_data_buffer_get_allocator(ZM_DATA_BUFFER  * data_buffer)
{
	return data_buffer -> vt -> get_allocator_fn(data_buffer -> opaque);
}

/*
 * Reset data_buffer
 */
void zm_data_buffer_reset(ZM_DATA_BUFFER  * data_buffer)
{
	data_buffer -> vt -> reset_fn(data_buffer -> opaque);
}

/*
 * Get data_buffer element
 */
char  * zm_data_buffer_get(const ZM_DATA_BUFFER  * data_buffer)
{
	return data_buffer -> vt -> get_fn(data_buffer -> opaque);
}

/*
 * Get data length
 */
int64_t zm_data_buffer_get_size(const ZM_DATA_BUFFER  * data_buffer)
{
	return data_buffer -> vt -> size_fn(data_buffer -> opaque);
}

/*
 * Get object capacity
 */
int64_t zm_data_buffer_get_capacity(const ZM_DATA_BUFFER  * data_buffer)
{
	return data_buffer -> vt -> size_fn(data_buffer -> opaque) +
	       data_buffer -> vt -> get_tail_size_fn(data_buffer -> opaque);
}

/*
 * Append other data buffer
 */
int64_t zm_data_buffer_append(ZM_DATA_BUFFER        * data_buffer,
                              const ZM_DATA_BUFFER  * data)
{
	;;
return zm_data_buffer_append_data(data_buffer, zm_data_buffer_get(data), zm_data_buffer_get_size(data));
}

/*
 * Append one character
 */
int64_t zm_data_buffer_append_char(ZM_DATA_BUFFER  * data_buffer,
                                   char              data)
{
	;;
return zm_data_buffer_append_data(data_buffer, &data, 1);
}

/*
 * Append asciz string
 */
int64_t zm_data_buffer_append_str(ZM_DATA_BUFFER  * data_buffer,
                                  const char      * data)
{
	;;
return zm_data_buffer_append_data(data_buffer, data, strlen(data));
}

/*
 * Append data
 */
int64_t zm_data_buffer_append_data(ZM_DATA_BUFFER  * data_buffer,
                                   const void      * data,
                                   uint64_t          size)
{
	;;
return data_buffer -> vt -> append_data_fn(data_buffer -> opaque, data, size);
}

/*
 * Assign other data buffer
 */
int64_t zm_data_buffer_assign(ZM_DATA_BUFFER        * data_buffer,
                              const ZM_DATA_BUFFER  * data)
{
	;;
return zm_data_buffer_assign_data(data_buffer, zm_data_buffer_get(data), zm_data_buffer_get_size(data));
}

/*
 * Assign one character
 */
int64_t zm_data_buffer_assign_char(ZM_DATA_BUFFER  * data_buffer,
                                   char              data)
{
	;;
return zm_data_buffer_assign_data(data_buffer, &data, 1);
}

/*
 * Assign asciz string
 */
int64_t zm_data_buffer_assign_str(ZM_DATA_BUFFER  * data_buffer,
                                  const char      * data)
{
	;;
return zm_data_buffer_assign_data(data_buffer, data, strlen(data));
}

/*
 * Assign data
 */
int64_t zm_data_buffer_assign_data(ZM_DATA_BUFFER  * data_buffer,
                                   const void      * data,
                                   uint64_t          size)
{
	;;
return data_buffer -> vt -> assign_data_fn(data_buffer -> opaque, data, size);
}

/*
 * Reserve space
 */
int64_t zm_data_buffer_reserve(ZM_DATA_BUFFER  * data_buffer,
                               uint64_t          size)
{
	;;
return data_buffer -> vt -> reserve_fn(data_buffer -> opaque, size);
}

/*
 * Set size
 */
int64_t zm_data_buffer_set_size(ZM_DATA_BUFFER  * data_buffer,
                                uint64_t          size)
{
	;;
return data_buffer -> vt -> set_size_fn(data_buffer -> opaque, size);
}

/*
 * Get free space length
 */
int64_t zm_data_buffer_get_tail_size(const ZM_DATA_BUFFER  * data_buffer)
{
	;;
return data_buffer -> vt -> get_tail_size_fn(data_buffer -> opaque);
}

/*
 * Get pointer
 */
char * zm_data_buffer_get_tail(const ZM_DATA_BUFFER  * data_buffer)
{
	;;
return data_buffer -> vt -> get_tail_fn(data_buffer -> opaque);
}
/* End */
