/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include <string.h>

#include "zm_data_buffer_base.h"
#include "zm_data_buffer_impl.h"
#include "zm_mem.h"

/**
 * @brief Create data buffer implementation object
 * @param alloc - Allocator object;
 */
static void * zm_data_buffer_impl_create(ZM_ALLOCATOR  * alloc);

/**
 * @brief Destroy data buffer
 * @param opaque - Data buffer object
 */
static void zm_data_buffer_impl_destroy(void  * opaque);

/**
 * @brief Get allocator
 * @param opaque - Data buffer object
 * @return Pointer to the memory allocator
 */
static ZM_ALLOCATOR * zm_data_buffer_impl_get_allocator(void  * opaque);

/**
 * @brief Reset data buffer
 * @param opaque - Data buffer object
 */
static void zm_data_buffer_impl_reset(void  * opaque);

/**
 * @brief Get data
 * @param opaque - Data buffer object
 * @return Pointer to the data buffer
 */
static char * zm_data_buffer_impl_get(const void  * opaque);

/**
 * @brief Get data length
 * @param opaque - Data buffer object
 * @return Data buffer size
 */
static int64_t zm_data_buffer_impl_get_size(const void  * opaque);

/**
 * @brief Append data
 * @param opaque - Data buffer object
 * @param data - Data to add
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
static int64_t zm_data_buffer_impl_append_data(void        * opaque,
                                               const char  * data,
                                               uint64_t      size);

/**
 * @brief Assign data
 * @param opaque - Data buffer object
 * @param data - Data to add
 * @param size - Data size
 * @return New data size or -1 if any error occured
 */
static int64_t zm_data_buffer_impl_assign_data(void        * opaque,
                                               const char  * data,
                                               uint64_t      size);

/**
 * @brief Reserve space
 * @param opaque - Data buffer object
 * @param size - New size
 * @return New data size or -1 if any error occured
 */
static int64_t zm_data_buffer_impl_reserve(void      * opaque,
                                           uint64_t    size);

/**
 * @brief Set size
 * @param opaque - Data buffer object
 * @return New data size or -1 if any error occured
 */
static int64_t zm_data_buffer_impl_set_size(void      * opaque,
                                            uint64_t    size);

/*
 * Get free space length
 */
static int64_t zm_data_buffer_impl_get_tail_size(void  * opaque);

/*
 * Get pointer
 */
static char * zm_data_buffer_impl_get_tail(void  * opaque);


/**
 * @struct _zm_data_buffer zm_data_buffer.h <zm_data_buffer.h>
 * @brief Data buffer, virtual methods table
 */
static ZM_DATA_BUFFER_VT zm_data_buffer_impl_vt =
{
	/** Create data buffer              */
	zm_data_buffer_impl_create,
	/** Destroy dat buffer             */
	zm_data_buffer_impl_destroy,
	/** Get allocator                   */
	zm_data_buffer_impl_get_allocator,
	/** Reset data buffer               */
	zm_data_buffer_impl_reset,
	/** Get data                        */
	zm_data_buffer_impl_get,
	/** Get data length                 */
	zm_data_buffer_impl_get_size,
	/** Append data                     */
	zm_data_buffer_impl_append_data,
	/** Assign data                     */
	zm_data_buffer_impl_assign_data,
	/** Reserve space                   */
	zm_data_buffer_impl_reserve,
	/** Set size                        */
	zm_data_buffer_impl_set_size,
	/** Get free space length           */
	zm_data_buffer_impl_get_tail_size,
	/** Get pointer                     */
	zm_data_buffer_impl_get_tail
};

/*
 * Create data_buffer
 */
ZM_DATA_BUFFER * zm_data_buffer_create_impl(ZM_DATA_BUFFER  * data_buffer,
                                            ZM_ALLOCATOR    * alloc)
{
	uint32_t  flags  = 0x00;

	if (data_buffer == NULL)
	{
		data_buffer = (ZM_DATA_BUFFER *)zm_allocate(alloc, sizeof(ZM_DATA_BUFFER));
		flags |= 0x01;
	}

	if (data_buffer == NULL) { return NULL; }

	data_buffer -> vt     = &zm_data_buffer_impl_vt;
	data_buffer -> flags  = flags;
	data_buffer -> opaque = zm_data_buffer_impl_create(alloc);

	if (data_buffer -> opaque == NULL)
	{
		if (data_buffer -> flags & 0x01) { zm_deallocate(alloc, data_buffer); }
		return NULL;
	}

return data_buffer;
}

/*
 * Create data_buffer
 */
void * zm_data_buffer_impl_create(ZM_ALLOCATOR  * alloc)
{
	ZM_DATA_BUFFER_IMPL * data_buffer_impl = (ZM_DATA_BUFFER_IMPL *)zm_allocate(alloc, sizeof(ZM_DATA_BUFFER_IMPL));
	if (data_buffer_impl == NULL) { return NULL; }

	data_buffer_impl -> size     = 0;
	data_buffer_impl -> capacity = 0;
	data_buffer_impl -> data     = NULL;
	data_buffer_impl -> alloc    = alloc;

return data_buffer_impl;
}

/*
 * Destroy data_buffer
 */
static void zm_data_buffer_impl_destroy(void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;
	ZM_ALLOCATOR         * alloc       = data_buffer -> alloc;

	zm_deallocate(alloc, data_buffer -> data);
	zm_deallocate(alloc, data_buffer);
}

/*
 * Get allocator
 */
static ZM_ALLOCATOR * zm_data_buffer_impl_get_allocator(void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;

return data_buffer -> alloc;
}

/*
 * Reset data_buffer
 */
static void zm_data_buffer_impl_reset(void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;

	data_buffer -> size = 0;
}

/*
 * Get data_buffer element
 */
static char  * zm_data_buffer_impl_get(const void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;
return data_buffer -> data;
}

/*
 * Get data length
 */
static int64_t zm_data_buffer_impl_get_size(const void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;

return data_buffer -> size;
}

/*
 * Append data
 */
static int64_t zm_data_buffer_impl_append_data(void        * opaque,
                                               const char  * data,
                                               uint64_t      size)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;
	ZM_ALLOCATOR         * alloc       = data_buffer -> alloc;

	if (data_buffer -> size + size > data_buffer -> capacity)
	{
		char      * new_data;
		uint64_t    new_capacity = data_buffer -> size + size;
		if (new_capacity < data_buffer -> capacity * 2) { new_capacity = data_buffer -> capacity * 2; }

		new_data = (char *)zm_reallocate(alloc, data_buffer -> data, new_capacity);

		if (new_data == NULL) { return -1; }

		data_buffer -> data     = new_data;
		data_buffer -> capacity = new_capacity;
	}

	memcpy(data_buffer -> data + data_buffer -> size, data, size);
	data_buffer -> size += size;

return data_buffer -> size;
}

/*
 * Assign data
 */
int64_t zm_data_buffer_impl_assign_data(void        * opaque,
                                        const char  * data,
                                        uint64_t      size)
{
	zm_data_buffer_impl_reset(opaque);

return zm_data_buffer_impl_append_data(opaque, data, size);
}

/*
 * Reserve space
 */
int64_t zm_data_buffer_impl_reserve(void      * opaque,
                                    uint64_t    size)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;
	ZM_ALLOCATOR         * alloc       = data_buffer -> alloc;

	if (data_buffer -> size + size > data_buffer -> capacity)
	{
		char      * new_data;
		uint64_t    new_capacity = data_buffer -> size + size;
		if (new_capacity < data_buffer -> capacity * 2) { new_capacity = data_buffer -> capacity * 2; }

		new_data = (char *)zm_reallocate(alloc, data_buffer -> data, new_capacity);
		if (new_data == NULL) { return -1; }

		data_buffer -> data     = new_data;
		data_buffer -> capacity = new_capacity;
	}

return (data_buffer -> capacity - data_buffer -> size);
}

/*
 * Set size
 */
int64_t zm_data_buffer_impl_set_size(void      * opaque,
                                     uint64_t    size)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;

	if (size > data_buffer -> capacity)
	{
		data_buffer -> size = 0;
		if (zm_data_buffer_impl_reserve(data_buffer, size) == -1) { return -1; }
		return size;
	}

	data_buffer-> size = size;
return size;
}

/*
 * Get free space length
 */
int64_t zm_data_buffer_impl_get_tail_size(void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;

return data_buffer -> capacity - data_buffer -> size;
}

/*
 * Get pointer
 */
char * zm_data_buffer_impl_get_tail(void  * opaque)
{
	ZM_DATA_BUFFER_IMPL  * data_buffer = (ZM_DATA_BUFFER_IMPL *)opaque;
return data_buffer -> data + data_buffer -> size;
}

/* End */
