/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_iterator.h"
#include "zm_helpers.h"
#include "zm_mem.h"

#include <strings.h>

/*
 * Skip space between tokens
 */
int32_t zm_skip_space(ZM_ITERATOR        * it_start,
                      const ZM_ITERATOR  * it_end,
                      const int32_t        ignore_lf)
{
	if (ignore_lf)
	{
		/* Skip spaces */
		for (;;)
		{
			/* End of buffer */
			if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }
			/* First NOT space */
			if (zm_iterator_ref(it_start) > ' ') { break; }
			zm_iterator_next(it_start);
		}
	}

	/* Skip spaces */
	for (;;)
	{
		/* End of buffer */
		if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }
		/* First NOT space */
		if (zm_iterator_ref(it_start) != ' ') { break; }
		/* Line feed */
		if (zm_iterator_ref(it_start) == 0x0A)  { return ZM_LF; }
		zm_iterator_next(it_start);
	}

return ZM_FOUND;
}

/*
 * Get "key[delimiter]"
 */
int32_t zm_get_key(ZM_ITERATOR          * it_start,
                   const ZM_ITERATOR    * it_end,
                   const unsigned char    delim,
                   ZM_ITERATOR          * it_key_start,
                   ZM_ITERATOR          * it_key_end,
                   const int32_t          ignore_lf)
{
	int32_t rc = 0;

	zm_iterator_copy(it_key_start, it_start);

	/* Get token */
	for (;;)
	{
		/* End of buffer */
		if (zm_iterator_equal(it_start, it_end))
		{
			zm_iterator_copy(it_key_end, it_start);
			return ZM_END;
		}

		/* Delimiter */
		if (zm_iterator_ref(it_start) == delim)
		{
			zm_iterator_copy(it_key_end, it_start);
			break;
		}

		if(zm_iterator_ref(it_start) == ' ')
		{
			zm_iterator_copy(it_key_end, it_start);

			rc = zm_skip_space(it_start, it_end, ignore_lf);
			if (rc != ZM_FOUND) { return rc; }

			if (zm_iterator_ref(it_start) != delim) { return ZM_FAILED; }
			break;
		}

		/* Line feed */
		if (!ignore_lf && (zm_iterator_ref(it_start) == 0x0D || zm_iterator_ref(it_start) == 0x0A)) { return ZM_FAILED; }

		zm_iterator_next(it_start);
	}
	zm_iterator_next(it_start);

return ZM_FOUND;
}

/*
 * Get "key[delimiter]"
 */
int32_t zm_get_key2(ZM_ITERATOR          * it_start,
                    const ZM_ITERATOR    * it_end,
                    const unsigned char    delim1,
                    const unsigned char    delim2,
                    ZM_ITERATOR          * it_key_start,
                    ZM_ITERATOR          * it_key_end,
                    const int32_t          ignore_lf)
{
	int32_t rc = 0;

	zm_iterator_copy(it_key_start, it_start);

	/* Get token */
	for (;;)
	{
		/* End of buffer */
		if (zm_iterator_equal(it_start, it_end))
		{
			zm_iterator_copy(it_key_end, it_start);
			return ZM_END;
		}

		/* Found delimiter? */
		if (zm_iterator_ref(it_start) == delim1 ||
		    zm_iterator_ref(it_start) == delim2)
		{
			zm_iterator_copy(it_key_end, it_start);
			return ZM_FOUND;
		}

		/* Remove spaces */
		if(zm_iterator_ref(it_start) == ' ')
		{
			zm_iterator_copy(it_key_end, it_start);

			rc = zm_skip_space(it_start, it_end, ignore_lf);
			if (rc != ZM_FOUND) { return rc; }

			/* Found delimiter? */
			if (zm_iterator_ref(it_start) == delim1 ||
			    zm_iterator_ref(it_start) == delim2)
			{
				return ZM_FOUND;
			}

			return ZM_FAILED;
		}
		zm_iterator_next(it_start);
	}

return ZM_FOUND;
}

/*
 * Parse "key: value" pairs
 */
int32_t zm_get_key_val(ZM_ITERATOR        * it_start,
                       const ZM_ITERATOR  * it_end,
                       ZM_ITERATOR        * it_key_start,
                       ZM_ITERATOR        * it_key_end,
                       ZM_ITERATOR        * it_val_start,
                       ZM_ITERATOR        * it_val_end,
                       uint64_t           * line)
{
	int32_t rc = 0;

	zm_iterator_copy(it_key_start, it_start);
	zm_iterator_copy(it_key_end,   it_start);
	zm_iterator_copy(it_val_start, it_start);
	zm_iterator_copy(it_val_end,   it_start);

	/* Skip spaces */
	rc = zm_skip_space(it_start, it_end, 0);
	if (rc != ZM_FOUND) { return rc; }

	/** Get key */
	rc = zm_get_key(it_start, it_end, ':', it_key_start, it_key_end, 0);
	if (rc != ZM_FOUND) { return rc; }

	/* Skip spaces */
	rc = zm_skip_space(it_start, it_end, 0);
	if (rc != ZM_FOUND) { return rc; }

	/* Start of value */
	zm_iterator_copy(it_val_start, it_start);
	zm_iterator_copy(it_val_end,   it_start);

	/* Get value */
	for (;;)
	{
		/* Get part of value */
		for (;;)
		{
			/* End of buffer */
			if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }

			/* Carriage return AND line feed */
			if (zm_iterator_ref(it_start) == 0x0D)
			{
				zm_iterator_copy(it_val_end, it_start);
				zm_iterator_next(it_start);

				/* Line feed */
				if (zm_iterator_equal(it_start, it_end) || zm_iterator_ref(it_start) != 0x0A) { return ZM_FAILED; }

				++(*line);
				break;
			}

			/* Line feed ONLY */
			if (zm_iterator_ref(it_start) == 0x0A)
			{
				zm_iterator_copy(it_val_end, it_start);

				++(*line);
				break;
			}

			zm_iterator_next(it_start);
		}
		zm_iterator_next(it_start);

		/* End of buffer */
		if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }

		/* Carriage return */
		if (zm_iterator_ref(it_start) == 0x0D)
		{
			zm_iterator_next(it_start);
			if (zm_iterator_equal(it_start, it_end)) { return ZM_FAILED; }
		}

		/* Line feed */
		if (zm_iterator_ref(it_start) == 0x0A)
		{
			++(*line);
			return ZM_LF;
		}

		/* Not a Tab or Space - stop processing */
		if (zm_iterator_ref(it_start) != 0x09 && zm_iterator_ref(it_start) != 0x20) { break; }
	}

return ZM_FOUND;
}

/*
 * Parse "x/y; foo=bar; bar=baz; baz=boo" string
 */
int32_t zm_parse_key_val(ZM_ITERATOR        * it_start,
                         const ZM_ITERATOR  * it_end,
                         ZM_BLOCK           * raw_key,
                         ZM_BLOCK           * raw_val,
                         uint32_t           * pair_num,
                         ZM_KEY_VAL        ** key_vals,
                         ZM_ALLOCATOR       * alloc,
                         const int32_t        special_case)
{
	int32_t rc = 0;

	ZM_KEY_VAL  * key_val = NULL;
	ZM_ITERATOR   token_start;
	ZM_ITERATOR   token_end;

	if (special_case)
	{
		/* Skip spaces  */
		rc = zm_skip_space(it_start, it_end, 1);
		if (rc != ZM_FOUND) { return rc; }

		/* Get key  */
		rc = zm_get_key2(it_start, it_end, '/', ';', &token_start, &token_end, 1);

		zm_block_create(raw_key, zm_iterator_get_pos(&token_start), zm_iterator_get_pos(&token_end));

		if (rc == ZM_END) { return rc; }

		if (zm_iterator_ref(it_start) == '/')
		{
			zm_iterator_next(it_start);
			rc = zm_get_key2(it_start, it_end, '/', ';', &token_start, &token_end, 1);

			zm_block_create(raw_val, zm_iterator_get_pos(&token_start), zm_iterator_get_pos(&token_end));

			if (rc == ZM_END) { return rc; }
		}
		zm_iterator_next(it_start);
	}

	for(;;)
	{
		if (key_val == NULL)
		{
			key_val = zm_allocate(alloc, sizeof(ZM_KEY_VAL));
			if (key_val == NULL) { return ZM_FATAL; }

			*key_vals = key_val;
		}
		else
		{
			key_val -> next = zm_allocate(alloc, sizeof(ZM_KEY_VAL));
			key_val = key_val -> next;
		}
		*pair_num += 1;

		/* Skip spaces */
		rc = zm_skip_space(it_start, it_end, 1);
		if (rc != ZM_FOUND) { return rc; }

		/* Get key */
		rc = zm_get_key(it_start, it_end, '=', &token_start, &token_end, 1);
		if (rc != ZM_FOUND) { return rc; }

		zm_block_create(&(key_val -> key), zm_iterator_get_pos(&token_start), zm_iterator_get_pos(&token_end));

		/* Skip spaces  */
		rc = zm_skip_space(it_start, it_end, 1);
		if (rc != ZM_FOUND) { return rc; }

		/* Search quoted token */
		if (zm_iterator_ref(it_start) == '"')
		{
			zm_iterator_next(it_start);
			zm_iterator_copy(&token_start, it_start);

			/* Get quoted token value */
			for (;;)
			{
				/* End of buffer */
				if (zm_iterator_equal(it_start, it_end)) { break; }
				/* Quote */
				if (zm_iterator_ref(it_start) == '"')   { break; }
				zm_iterator_next(it_start);
			}

			zm_iterator_copy(&token_end, it_start);
			zm_iterator_next(it_start);

			zm_block_create(&(key_val -> val), zm_iterator_get_pos(&token_start), zm_iterator_get_pos(&token_end));

			/* Skip spaces */
			rc = zm_skip_space(it_start, it_end, 1);
			if (rc != ZM_FOUND) { return rc; }

			if (zm_iterator_ref(it_start) != ';') { return ZM_FAILED; }
			zm_iterator_next(it_start);
		}
		else
		{
			/* Get key */
			rc = zm_get_key(it_start, it_end, ';', &token_start, &token_end, 1);

			zm_block_create(&(key_val -> val), zm_iterator_get_pos(&token_start), zm_iterator_get_pos(&token_end));

			if (rc != ZM_FOUND) { return rc; }
		}
	}

return ZM_FOUND;
}

/*
 * Compare key with given pattern
 */
int32_t zm_compare(const char      * key,
                   const uint64_t    key_len,
                   const char      * pattern)
{
	return strncasecmp((const char *)key, pattern, key_len);
}

/*
 * Compare strings
 */
uint32_t zm_compare_list(const char      * key,
                         const uint64_t    key_len,
                         const char     ** keys,
                         const uint32_t  * vals)
{
	while (*keys != NULL)
	{
		if (zm_compare(key, key_len, *keys) == 0) { return *vals; }
		++keys; ++vals;
	}
return *vals;
}

/*
 * Find substring
 */
int32_t zm_find(ZM_ITERATOR          * it_start,
                const unsigned char  * sequence,
                uint64_t               seq_len,
                uint64_t             * line)
{
	uint64_t cur_line = *line;

	while (seq_len > 0)
	{
		if (zm_iterator_ref(it_start) != *sequence) { return 0; }

		/* Line feed */
		if (zm_iterator_ref(it_start) == 0x0A) { ++cur_line; }

		zm_iterator_next(it_start);
		++sequence; --seq_len;
	}

	*line = cur_line;
return 1;
}

/*
 * Find boundary
 */
int32_t zm_find_boundary(ZM_ITERATOR          * it_start,
                         ZM_ITERATOR          * it_boundary_start,
                         ZM_ITERATOR          * it_end,
                         const unsigned char  * str_boundary,
                         const uint64_t         boundary_len,
                         uint64_t             * line,
                         const int32_t          skip_crlf)
{
	int32_t cr_flag = 0;
	uint64_t cur_line = *line;
	zm_iterator_copy(it_boundary_start, it_start);

	if (skip_crlf) { goto FIRST_CRLF; }
	for (;;)
	{
		uint64_t distance = 0;
		ZM_ITERATOR it_tmp;

		cr_flag = 0;
		/* End of buffer */
		if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }

		/* CR? */
		if (zm_iterator_ref(it_start) == 0x0D)
		{
			zm_iterator_copy(it_boundary_start, it_start);

			zm_iterator_next(it_start);
			if (zm_iterator_equal(it_start, it_end)) { return ZM_FAILED; }

			cr_flag = 1;
		}

		/* LF */
		if (zm_iterator_ref(it_start) != 0x0A && cr_flag == 1) { return ZM_FAILED; }

		if (cr_flag == 0) { zm_iterator_copy(it_boundary_start, it_start); }

		if (zm_iterator_ref(it_start) == 0x0A) { ++cur_line; }
		zm_iterator_next(it_start);
FIRST_CRLF:
		zm_iterator_copy(&it_tmp, it_start);
		distance = zm_iterator_get_distance(it_start, it_end);

		/* --BOUNDARY */
		if (distance < boundary_len + 2) { return ZM_END; }
		if (zm_iterator_ref(it_start) == '-')
		{
			zm_iterator_next(it_start);

			if (zm_iterator_ref(it_start) =='-')
			{
				zm_iterator_next(it_start);

				if (zm_find(it_start, str_boundary, boundary_len, &cur_line))
				{
					int32_t tail_len = (cr_flag == 1) ? 4 : 2;
					/* --BOUNDARY CR? LF */
					if (distance < boundary_len + tail_len) { return ZM_FAILED; }

					/* --BOUNDARY-- CR? LF */
					if (zm_iterator_ref(it_start) == '-')
					{
						if (distance < boundary_len + tail_len + 2) { return ZM_FAILED; }
						zm_iterator_next(it_start);

						if (zm_iterator_ref(it_start) != '-') { return ZM_FAILED; }
						zm_iterator_next(it_start);

						/* CR? */
						if (zm_iterator_ref(it_start) == 0x0D)
						{
							zm_iterator_next(it_start);
						}
						else if (cr_flag == 1)
						{
							return ZM_FAILED;
						}

						/* LF */
						if (zm_iterator_ref(it_start) != 0x0A) { return ZM_FAILED; }
						zm_iterator_next(it_start);

						*line = ++cur_line;
						return ZM_BOUNDARY_END;
					}

					/* CR? */
					if (zm_iterator_ref(it_start) == 0x0D)
					{
						zm_iterator_next(it_start);
					}
					else if (cr_flag == 1)
					{
						return ZM_FAILED;
					}

					/* LF */
					if (zm_iterator_ref(it_start) != 0x0A) { return ZM_FAILED; }
					zm_iterator_next(it_start);

					*line = ++cur_line;
					return ZM_BOUNDARY;
				}
			}
			zm_iterator_copy(it_start, &it_tmp);
		}
	}
return ZM_FAILED;
}

/* End. */
