/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_data_buffer.h"
#include "zm_iconv.h"
#include "zm_mem.h"
#include "zm_rbtree_cmp.h"
#include "zm_rbtree_fn.h"

#include <iconv.h>
#include <errno.h>
#include <string.h>

#ifndef ZM_ICONV_BUFFER_LEN
#define ZM_ICONV_BUFFER_LEN 4096
#endif /* ZM_ICONV_BUFFER_LEN */

/*
 * Deallocate charset map
 */
static void zm_iconv_destroy_map(void  * opaque,
                                 void  * key,
                                 void  * val)
{
	ZM_ALLOCATOR  * alloc = (ZM_ALLOCATOR *)opaque;

	zm_deallocate(alloc, key);
	zm_deallocate(alloc, val);
}

/*
 * Get iconv object
 */
static iconv_t zm_iconv_get(ZM_ICONV    * zm_iconv,
                            const char  * src_charset,
                            const char  * dst_charset)
{
	iconv_t obj = iconv_open(dst_charset, src_charset);
	/* Check aliases */
	if (obj == (iconv_t)(-1))
	{
		const char * alias_src_charset = zm_rb_search(&(zm_iconv -> encodings_map), src_charset) -> val;
		const char * alias_dst_charset = zm_rb_search(&(zm_iconv -> encodings_map), dst_charset) -> val;

		/* Nothing to do */
		if (alias_src_charset == ((void *)-1) && alias_dst_charset == ((void *)-1)) { return (iconv_t)(-1); }

		if (alias_src_charset == ((void *)-1)) { alias_src_charset = src_charset; }
		if (alias_dst_charset == ((void *)-1)) { alias_dst_charset = dst_charset; }

		obj = iconv_open(alias_dst_charset, alias_src_charset);
	}

return obj;
}

/*
 * Create text converter
 */
ZM_ICONV * zm_iconv_create(ZM_ICONV      * zm_iconv,
                           ZM_ALLOCATOR  * alloc)
{
	uint32_t flags = 0;

	if (zm_iconv == NULL)
	{
		zm_iconv = (ZM_ICONV *)zm_allocate(alloc, sizeof(ZM_ICONV));
		flags = 1;
	}

	if (zm_iconv == NULL) { return NULL; }

	zm_iconv -> flags     = flags;
	zm_iconv -> alloc = alloc;

	zm_rb_create(&(zm_iconv -> encodings_map), alloc, zm_rb_string_case_cmp, zm_iconv_destroy_map, alloc);

return zm_iconv;
}

/*
 * Add encoding alias
 */
int32_t zm_iconv_add_alias(ZM_ICONV    * zm_iconv,
                           const char  * encoding,
                           const char  * alias)
{
	ZM_ALLOCATOR  * alloc = zm_iconv -> alloc;

	char      * key = zm_duplicate(alloc, alias, strlen(alias) + 1);
	char      * val = zm_duplicate(alloc, encoding, strlen(encoding) + 1);

	if (key == NULL || val == NULL)
	{
		zm_deallocate(alloc, key);
		zm_deallocate(alloc, val);
		return -1;
	}

	if (zm_rb_insert(&(zm_iconv -> encodings_map), key, val) == -1)
	{
		zm_deallocate(alloc, key);
		zm_deallocate(alloc, val);
		return -1;
	}

return 0;
}

/*
 * Remove encoding alias
 */
int32_t zm_iconv_remove_alias(ZM_ICONV    * zm_iconv,
                              const char  * alias)
{
	;;
return zm_rb_delete(&(zm_iconv -> encodings_map), alias);
}

/*
 * Append data
 */
static void zm_iconv_append(ZM_ICONV        * zm_iconv,
                            void           ** result,
                            uint64_t        * result_len,
                            const void      * src,
                            const uint64_t    src_len)
{
	ZM_ALLOCATOR  * alloc = zm_iconv -> alloc;

	size_t  tmp_size = *result_len + src_len;
	void  * tmp      = zm_reallocate(alloc, *result, tmp_size + 1);

	if (tmp == NULL)
	{
		zm_deallocate(alloc, *result);
		*result_len = 0;
		return;
	}

	memcpy((char *)tmp + *result_len, src, src_len);

	*result      = tmp;
	*result_len  = tmp_size;
}

/*
 * Convert text
 */
int32_t zm_iconv_convert(ZM_ICONV        * zm_iconv,
                         const void      * src,
                         const uint64_t    src_len,
                         void           ** result,
                         uint64_t        * result_len,
                         const char      * src_charset,
                         const char      * dst_charset,
                         uint32_t        * errors,
                         const int32_t     ignore_ilseq)
{
	char          buf[ZM_ICONV_BUFFER_LEN];
	size_t        src_iconv_len = src_len;

#if defined(linux) || defined(__linux__) || defined(__APPLE__) || defined(sun) || defined(__FreeBSD__)
	char        * src_iconv = (char *)src;
#else
	const char  * src_iconv = (char *)src;
#endif

	iconv_t obj = zm_iconv_get(zm_iconv, src_charset, dst_charset);
	if (obj == (iconv_t)(-1)) { return -1; }

	*errors = 0;
	*result_len = 0;
	for (;;)
	{
		char    * dst_iconv     = buf;
		size_t    dst_iconv_len = ZM_ICONV_BUFFER_LEN;
		size_t    rc            = iconv(obj, &src_iconv, &src_iconv_len, &dst_iconv, &dst_iconv_len);
		size_t    converted     = dst_iconv - buf;

		if (converted > 0)
		{
			zm_iconv_append(zm_iconv, result, result_len, buf, converted);
			if (result == NULL)
			{
				iconv_close(obj);
				return -1;
			}
		}

		/* All data converted? */
		if (rc != (size_t)-1) { break; }

		if (errno != E2BIG)
		{
			*errors += 1;
			if (!ignore_ilseq)
			{
				zm_iconv -> alloc -> deallocate(zm_iconv -> alloc, *result);
				*result     = NULL;
				*result_len = 0;

				iconv_close(obj);
				return -1;
			}
			++src_iconv;
			--src_iconv_len;
		}
	}

	/* Make asciz */
	((char *)(*result))[*result_len] = 0;

	iconv_close(obj);
return 0;
}

/**
 * @brief Convert text
 */
int32_t zm_iconv_convert_data_buffer(ZM_ICONV        * zm_iconv,
                                     const void      * src,
                                     const uint64_t    src_len,
                                     ZM_DATA_BUFFER  * result,
                                     const char      * src_charset,
                                     const char      * dst_charset,
                                     uint32_t        * errors,
                                     const int32_t     ignore_ilseq)
{
	char          buf[ZM_ICONV_BUFFER_LEN];
	size_t        src_iconv_len = src_len;

#if defined(linux) || defined(__linux__) || defined(__APPLE__) || defined(sun) || defined(__FreeBSD__)
	char        * src_iconv = (char *)src;
#else
	const char  * src_iconv = (char *)src;
#endif

	iconv_t obj = zm_iconv_get(zm_iconv, src_charset, dst_charset);
	if (obj == (iconv_t)(-1)) { return -1; }

	*errors = 0;
	for (;;)
	{
		char    * dst_iconv     = buf;
		size_t    dst_iconv_len = ZM_ICONV_BUFFER_LEN;
		size_t    rc            = iconv(obj, &src_iconv, &src_iconv_len, &dst_iconv, &dst_iconv_len);
		size_t    converted     = dst_iconv - buf;

		if (converted > 0)
		{
			zm_data_buffer_append_data(result, buf, converted);
			if (result == NULL)
			{
				iconv_close(obj);
				return -1;
			}
		}

		/* All data converted? */
		if (rc != (size_t)-1) { break; }

		if (errno != E2BIG)
		{
			*errors += 1;

			if (!ignore_ilseq)
			{
				iconv_close(obj);
				return -1;
			}
			++src_iconv;
			--src_iconv_len;
		}
	}

	iconv_close(obj);
return 0;
}

/**
 * @brief Convert text
 */
int32_t zm_iconv_convert_data_buffer2(ZM_ICONV        * zm_iconv,
                                      ZM_DATA_BUFFER  * src,
                                      ZM_DATA_BUFFER  * result,
                                      const char      * src_charset,
                                      const char      * dst_charset,
                                      uint32_t        * errors,
                                      const int32_t     ignore_ilseq)
{
	;;
return zm_iconv_convert_data_buffer(zm_iconv, zm_data_buffer_get(src), zm_data_buffer_get_size(src), result, src_charset, dst_charset, errors, ignore_ilseq);
}

/*
 * Destroy text converter
 */
void zm_iconv_destroy(ZM_ICONV * zm_iconv)
{
	ZM_ALLOCATOR  * alloc = zm_iconv -> alloc;

	zm_rb_destroy(&(zm_iconv -> encodings_map));

	if (zm_iconv -> flags == 1)
	{
		zm_deallocate(alloc, zm_iconv);
	}
}
/* End. */
