/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_aliases.h"
#include "zm_helpers.h"
#include "zm_iterator.h"
#include "zm_mem.h"
#include "zm_parser.h"

#include <string.h>
#include <strings.h>

static const char                         * zm_cte_str[]                  = { "7bit", "8bit", "binary", "quoted-printable", "base64", "x-uuencode", NULL           };
static enum ZM_CONTENT_TRANSFER_ENCODING    zm_cte[]                      = {  BIT7,   BIT8,   BINARY,   QUOTED_PRINTABLE,   BASE64,   UU_ENCODED,  OTHER_ENCODING };

static const char                         * zm_types_str[]                = { "multipart", "text", "message", "image", "audio", "video", "application", "binary",     NULL       };
static enum ZM_CONTENT_TYPE                 zm_types[]                    = {  MULTIPART,   TEXT,   MESSAGE,   IMAGE,   AUDIO,   VIDEO,   APPLICATION,   BINARY_PART, OTHER_TYPE };

static const char                         * zm_basic_subtypes_str[]       = { "plain", "html", "xml", "rtf", "calendar", "directory", "rfc822-headers", "gif", "jpeg", "png", "svg", "tiff", "basic", "midi", "mpeg", "mp4", "ogg", "gzip", "pdf", "x-rar", "x-tar", "zip", "octet-stream", NULL          };
static enum ZM_CONTENT_SUBTYPE              zm_basic_subtypes[]           = {  PLAIN,   HTML,   XML,   RTF,   CALENDAR,   DIRECTORY,   RFC822_HEADERS,   GIF,   JPEG,   PNG,   SVG,   TIFF,   BASIC,   MIDI,   MPEG,   MP4,   OGG,   GZIP,   PDF,   X_RAR,   X_TAR,   ZIP,   OCTET_STREAM,  OTHER_SUBTYPE };

static const char                         * zm_multipart_subtypes_str[]   = { "mixed", "alternative", "related", "report", "parallel", "digest", "signed", "encrypted",  NULL          };
static enum ZM_CONTENT_SUBTYPE              zm_multipart_subtypes[]       = {  MIXED,   ALTERNATIVE,   RELATED,   REPORT,   PARALLEL,   DIGEST,   SIGNED,   ENCRYPTED,   OTHER_SUBTYPE };

static const char                         * zm_message_subtypes_str[]     = { "rfc822", "partial", "delivery-status", "disposition-notification-to", "external-body", "feedback-report", NULL          };
static enum ZM_CONTENT_SUBTYPE              zm_message_subtypes[]         = {  RFC822,   PARTIAL,   DELIVERY_STATUS,   DISPOSITION_NOTIFICATION,      EXTERNAL_BODY,   FEEDBACK_REPORT,  OTHER_SUBTYPE };

static const char                         * zm_content_dispositions_str[] = { "inline", "attachment", NULL   };
static enum ZM_CONTENT_DISPOSITION          zm_content_dispositions[]     = {  INLINE,   ATTACHMENT,  INLINE };

static const char     * zm_charsets_str[] = {
	"US-ASCII",

	"UTF-7",         "UTF_8",         "UTF_16",
	"KOI8-R",        "KOI8-U",
	"WINDOWS-1250",  "WINDOWS-1251",  "WINDOWS-1252",  "WINDOWS-1253",
	"WINDOWS-1254",  "WINDOWS-1255",  "WINDOWS-1256",  "WINDOWS-1257",
	"WINDOWS-1258",
	"ISO-8859-1",    "ISO-8859-2",    "ISO-8859-3",    "ISO-8859-4",
	"ISO-8859-5",    "ISO-8859-6",    "ISO-8859-7",    "ISO-8859-8",
	"ISO-8859-9",    "ISO-8859-10",   "ISO-8859-11",   "ISO-8859-13",
	"ISO-8859-14",   "ISO-8859-15",   "ISO-8859-16",
	"CP437",         "CP737",         "CP775",         "CP850",
	"CP852",         "CP853",         "CP855",         "CP857",
	"CP858",         "CP860",         "CP861",         "CP862",
	"CP863",         "CP864",         "CP865",         "CP866",
	"CP869",         "CP874",         "CP936",         "CP950",
	"CP1125",
	"TIS",           "SHIFT-JIS",     "BIG5",          "EUC-KR",
	NULL
};

static enum ZM_CHARSET  zm_charsets[] = {
	US_ASCII,
	UTF_7,         UTF_8,         UTF_16,
	KOI_8R,        KOI_8U,
	WINDOWS_1250,  WINDOWS_1251,  WINDOWS_1252,  WINDOWS_1253,
	WINDOWS_1254,  WINDOWS_1255,  WINDOWS_1256,  WINDOWS_1257,
	WINDOWS_1258,
	ISO8859_1,     ISO8859_2,     ISO8859_3,     ISO8859_4,
	ISO8859_5,     ISO8859_6,     ISO8859_7,     ISO8859_8,
	ISO8859_9,     ISO8859_10,    ISO8859_11,    ISO8859_13,
	ISO8859_14,    ISO8859_15,    ISO8859_16,
	CP_437,        CP_737,        CP_775,        CP_850,
	CP_852,        CP_853,        CP_855,        CP_857,
	CP_858,        CP_860,        CP_861,        CP_862,
	CP_863,        CP_864,        CP_865,        CP_866,
	CP_869,        CP_874,        CP_936,        CP_950,
	CP_1125,
	TIS,           SHIFT_JIS,     BIG5,          EUC_KR,
	OTHER_CHARSET
};

/*
 * Parse message
 */
static int32_t zm_do_parse_mime(const void     * msg,
                                ZM_ITERATOR    * it_start,
                                ZM_ITERATOR    * it_part_end,
                                ZM_ITERATOR    * it_end,
                                ZM_BLOCK       * boundary,
                                ZM_PART        * parent,
                                ZM_PART       ** part,
                                uint64_t       * line,
                                ZM_ALLOCATOR   * alloc,
                                ZM_ALIASES     * charset_aliases,
                                int32_t          ignore_errors);

/*
 * Parse headers
 */
static int32_t zm_parse_headers(ZM_ITERATOR    * it_start,
                                ZM_ITERATOR    * it_end,
                                ZM_HEADER      * header,
                                ZM_ALLOCATOR   * alloc,
                                uint64_t       * line)
{
	ZM_FIELD_REC  * field_rec = NULL;

	/* End of buffer */
	if (zm_iterator_equal(it_start, it_end)) { return ZM_END; }

	/* Carriage return */
	if (zm_iterator_ref(it_start) == 0x0D)
	{
		zm_iterator_next(it_start);
		if (zm_iterator_equal(it_start, it_end)) { return ZM_FAILED; }
	}

	/* Line feed */
	if (zm_iterator_ref(it_start) == 0x0A) { *line += 1; return ZM_LF; }

	for(;;)
	{
		int32_t rc = 0;

		ZM_ITERATOR     it_key_start;
		ZM_ITERATOR     it_key_end;
		ZM_ITERATOR     it_val_start;
		ZM_ITERATOR     it_val_end;
		ZM_FIELD_REC  * tmp = zm_allocate(alloc, sizeof(ZM_FIELD_REC));

		if (tmp == NULL) { return ZM_FATAL; }

		if (field_rec == NULL)
		{
			header -> fields = field_rec = tmp;
		}
		else
		{
			field_rec -> next = tmp;
			field_rec = tmp;
		}

		zm_iterator_zero(&it_key_start);
		zm_iterator_zero(&it_key_end);
		zm_iterator_zero(&it_val_start);
		zm_iterator_zero(&it_val_end);

		rc = zm_get_key_val(it_start, it_end, &it_key_start, &it_key_end, &it_val_start, &it_val_end, line);

		/* Key */
		zm_block_create(&(field_rec -> name), zm_iterator_get_pos(&it_key_start), zm_iterator_get_pos(&it_key_end));

		/* Value */
		zm_block_create(&(field_rec -> value.raw), zm_iterator_get_pos(&it_val_start), zm_iterator_get_pos(&it_val_end));

		header -> fields_num += 1;

		if (rc != ZM_FOUND) { return rc; }
	}

return ZM_END;
}

/*
 * Parse tail of message
 */
static void zm_parse_tail(ZM_ITERATOR   * it_start,
                          ZM_ITERATOR   * it_part_end,
                          ZM_ITERATOR   * it_end,
                          ZM_PART      ** part,
                          uint64_t        lf_count,
                          uint64_t      * line)
{
	/* Count number of lines */
	while (zm_iterator_not_equal(it_start, it_end))
	{
		if (zm_iterator_ref(it_start) == 0x0A) { *line += 1; }
		zm_iterator_next(it_start);
	}

	/* End of part */
	(*part) -> data.raw.end = zm_iterator_get_pos(it_start);
	(*part) -> data.lines   = *line - lf_count;

	zm_iterator_copy(it_part_end, it_start);
}

/*
 * Parse multipart message
 */
static int32_t zm_parse_multipart(const void     * msg,
                                  ZM_ITERATOR    * it_start,
                                  ZM_ITERATOR    * it_part_end,
                                  ZM_ITERATOR    * it_end,
                                  ZM_BLOCK       * boundary,
                                  ZM_PART        * parent,
                                  ZM_PART       ** part,
                                  uint64_t       * line,
                                  ZM_ALLOCATOR   * alloc,
                                  ZM_ALIASES     * charset_aliases,
                                  int32_t          ignore_errors)
{
	uint32_t          part_num = 1;
	int32_t           rc       = 0;
	const uint64_t    lf_count = *line;
	ZM_PART         * mpart    = *part;
	ZM_PART         * subpart  = NULL;

	uint64_t               boundary_len = 0;
	const unsigned char  * str_boundary = NULL;

	/* Start of part */
	mpart -> data.raw.start = mpart -> data.raw.end = zm_iterator_get_pos(it_start);

	/* Get boundary */
	boundary_len = 0;
	str_boundary = zm_block_get_data(&(mpart -> u.multi.boundary), msg, &boundary_len);
	if (boundary_len == 0) { return ZM_FAILED; }

	/* Skip preamble */
	rc = zm_find_boundary(it_start, it_part_end, it_end, str_boundary, boundary_len, line, 1);
	if (rc != ZM_BOUNDARY && rc != ZM_BOUNDARY_END) { return rc; }

	/* Parse subparts */
	while (rc != ZM_BOUNDARY_END)
	{
		ZM_PART * tmp = NULL;
		rc = zm_do_parse_mime(msg, it_start, it_part_end, it_end, &((*part) -> u.multi.boundary), mpart, &tmp, line, alloc, charset_aliases, ignore_errors);

		/* Error? */
		if (rc != ZM_BOUNDARY && rc != ZM_BOUNDARY_END)
		{
			if (tmp != NULL)
			{
				if (ignore_errors)
				{
					if (mpart -> u.multi.parts == NULL)
					{
						mpart -> u.multi.parts = subpart = tmp;
					}
					else
					{
						subpart -> next = tmp;
						subpart = tmp;
					}
				}
				else
				{
					zm_deallocate(alloc, tmp);
				}
			}

			return rc;
		}

		/* Store part number */
		tmp -> part_num = part_num;

		if (mpart -> u.multi.parts == NULL)
		{
			mpart -> u.multi.parts = subpart = tmp;
		}
		else
		{
			subpart -> next = tmp;
			subpart = tmp;
		}
		++part_num;
	}

	/* No boundary - parse to the end of data */
	if (boundary -> start == boundary -> end)
	{
		zm_parse_tail(it_start, it_part_end, it_end, part, lf_count, line);
		return ZM_END;
	}

	/* Get boundary */
	boundary_len = 0;
	str_boundary = zm_block_get_data(boundary, msg, &boundary_len);

	/* Skip epilogue */
	rc = zm_find_boundary(it_start, it_part_end, it_end, str_boundary, boundary_len, line, 1);
	if (rc != ZM_BOUNDARY && rc != ZM_BOUNDARY_END) { return rc; }

	/* End of part */
	mpart -> data.raw.end = zm_iterator_get_pos(it_part_end);
	mpart -> data.lines   = *line - lf_count - 2;

return rc;
}

/*
 * Parse message
 */
static int32_t zm_parse_message(const void     * msg,
                                ZM_ITERATOR    * it_start,
                                ZM_ITERATOR    * it_part_end,
                                ZM_ITERATOR    * it_end,
                                ZM_BLOCK       * boundary,
                                ZM_PART        * parent,
                                ZM_PART       ** part,
                                uint64_t       * line,
                                ZM_ALLOCATOR   * alloc,
                                ZM_ALIASES     * charset_aliases,
                                int32_t          ignore_errors)
{
	int32_t           rc       = 0;
	const uint64_t    lf_count = *line;
	ZM_PART         * mpart    = *part;

	/* Start of part */
	mpart -> data.raw.start = (*part) -> data.raw.end = zm_iterator_get_pos(it_start);

	/* Parse Message part */
	rc = zm_do_parse_mime(msg, it_start, it_part_end, it_end, boundary, mpart, &(mpart -> u.message.part), line, alloc, charset_aliases, ignore_errors);

	/* Store part number */
	if (mpart -> u.message.part != NULL)
	{
		mpart -> u.message.part -> part_num = mpart -> part_num;
		mpart -> u.message.part -> is_text  = 1;
	}

	/* End of part */
	mpart -> data.raw.end = zm_iterator_get_pos(it_part_end);
	mpart -> data.lines   = *line - lf_count - 2;

return rc;
}

/*
 * Parse text or basic part
 */
static int32_t zm_parse_text_or_basic(const void    * msg,
                                      ZM_ITERATOR   * it_start,
                                      ZM_ITERATOR   * it_part_end,
                                      ZM_ITERATOR   * it_end,
                                      ZM_BLOCK      * boundary,
                                      ZM_PART      ** part,
                                      uint64_t      * line)
{
	int32_t           rc       = 0;
	const uint64_t    lf_count = *line;
	ZM_PART         * mpart    = *part;

	uint64_t               boundary_len = 0;
	const unsigned char  * str_boundary = NULL;

	/* Start of part */
	mpart -> data.raw.start = mpart -> data.raw.end = zm_iterator_get_pos(it_start);

	/* No boundary - parse to the end of data */
	if (boundary -> start == boundary -> end)
	{
		zm_parse_tail(it_start, it_part_end, it_end, part, lf_count, line);
		return ZM_END;
	}

	/* Get boundary */
	str_boundary = zm_block_get_data(boundary, msg, &boundary_len);

	/* Find boundary */
	rc = zm_find_boundary(it_start, it_part_end, it_end, str_boundary, boundary_len, line, 1);

	/* End of part */
	mpart -> data.raw.end = zm_iterator_get_pos(it_part_end);
	mpart -> data.lines   = *line - lf_count - 2;

return rc;
}

/*
 * Copy block
 */
static void zm_copy_block(ZM_BLOCK  * bl_dst,
                          ZM_BLOCK  * bl_src)
{
	bl_dst -> start = bl_src -> start;
	bl_dst -> end   = bl_src -> end;
}

/*
 * Parse message
 */
static int32_t zm_do_parse_mime(const void     * msg,
                                ZM_ITERATOR    * it_start,
                                ZM_ITERATOR    * it_part_end,
                                ZM_ITERATOR    * it_end,
                                ZM_BLOCK       * boundary,
                                ZM_PART        * parent,
                                ZM_PART       ** part,
                                uint64_t       * line,
                                ZM_ALLOCATOR   * alloc,
                                ZM_ALIASES     * charset_aliases,
                                int32_t          ignore_errors)
{
	uint64_t      tmp_len = 0;
	const char  * tmp     = NULL;

	int32_t         rc                  = 0;
	ZM_PART       * mpart               = zm_allocate(alloc, sizeof(ZM_PART));
	ZM_HEADER     * header              = NULL;
	ZM_FIELD_REC  * content_type        = NULL;
	ZM_FIELD_REC  * content_disposition = NULL;
	ZM_FIELD_REC  * field_rec           = NULL;

	if (mpart == NULL) { return ZM_FATAL; }

	mpart -> alloc = alloc;

	*part = mpart;

	/* Store parent part */
	mpart -> parent   = parent;
	mpart -> is_text  = 0;
	mpart -> is_basic = 0;

	header = &(mpart -> header);

	header -> raw.start = zm_iterator_get_pos(it_start);
	rc = zm_parse_headers(it_start, it_end, header, alloc, line);

	/* Need line feed after headers */
	if (rc != ZM_LF)
	{
		header -> raw.end = zm_iterator_get_pos(it_start);
		return zm_parse_text_or_basic(msg, it_start, it_part_end, it_end, boundary, part, line);
	}

	/* Include second LF */
	zm_iterator_next(it_start);
	header -> raw.end = zm_iterator_get_pos(it_start);

	field_rec = header -> fields;
	while (field_rec != NULL)
	{
		/* Get key */
		uint64_t      key_len = 0;
		const char  * key     = (const char *)zm_block_get_data(&(field_rec -> name), msg, &key_len);

		/* Specical case for "Content-" headers */
		const uint64_t content_len = sizeof("content-") - 1;
		if (key_len > content_len && zm_compare((const char *)key, content_len, "content-") == 0)
		{
			ZM_ITERATOR  it_val_start;
			ZM_ITERATOR  it_val_end;

			const char      * content_val     = key     + content_len;
			const uint64_t    content_val_len = key_len - content_len;

			if (zm_compare(content_val, content_val_len, "type") == 0)
			{
				zm_iterator_create(&it_val_start, msg, field_rec -> value.raw.start);
				zm_iterator_create(&it_val_end,   msg, field_rec -> value.raw.end);

				content_type = field_rec;

				zm_parse_key_val(&it_val_start, &it_val_end, &(mpart -> raw_type), &(mpart -> raw_subtype), &(field_rec -> value.pair_num), &(field_rec ->  value.pairs), alloc, 1);
			}
			else if (strncasecmp(content_val, "disposition", content_val_len) == 0)
			{
				ZM_BLOCK      tmp;
				ZM_KEY_VAL  * kv;

				zm_iterator_create(&it_val_start, msg, field_rec -> value.raw.start);
				zm_iterator_create(&it_val_end,   msg, field_rec -> value.raw.end);

				content_disposition = field_rec;

				zm_parse_key_val(&it_val_start, &it_val_end, &(mpart -> raw_content_disposition.raw), &tmp, &(field_rec -> value.pair_num), &(field_rec -> value.pairs), alloc, 1);

				kv = field_rec -> value.pairs;
				while (kv != NULL)
				{
					uint64_t  key_len = 0;
					const unsigned char * key = zm_block_get_data(&(kv -> key), msg, &key_len);

					if (zm_compare((const char *)key, key_len, "filename") == 0)
					{
						mpart -> content_disposition_filename = &(kv -> val);
						break;
					}
					kv = kv -> next;
				}
			}
			else if (strncasecmp(content_val, "transfer-encoding", content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_transfer_encoding), &(field_rec -> value.raw)); }
			else if (strncasecmp(content_val, "language",          content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_language),          &(field_rec -> value.raw)); }
			else if (strncasecmp(content_val, "location",          content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_location),          &(field_rec -> value.raw)); }
			else if (strncasecmp(content_val, "id",                content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_id),                &(field_rec -> value.raw)); }
			else if (strncasecmp(content_val, "description",       content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_description),       &(field_rec -> value.raw)); }
			else if (strncasecmp(content_val, "md5",               content_val_len) == 0) { zm_copy_block(&(mpart -> raw_content_md5),               &(field_rec -> value.raw)); }
		}
		else if (strncasecmp(key, "message-id", key_len) == 0) { zm_copy_block(&(mpart -> raw_message_id), &(field_rec -> value.raw)); }

		field_rec = field_rec -> next;
	}

	/* Not found?
	 * Switch to default: text/plain; charset=US-ASCII
	 */
	if (content_type == NULL)
	{
		mpart -> type    = TEXT;
		mpart -> subtype = PLAIN;

		mpart -> content_type_val = NULL;

		mpart -> u.text.charset = US_ASCII;
	}
	else
	{
		uint64_t      content_type_len = 0;
		const char  * content_type_str = (const char *)zm_block_get_data(&(mpart -> raw_type), msg, &content_type_len);

		uint64_t       content_subtype_len = 0;
		const char  *  content_subtype_str = (const char *)zm_block_get_data(&(mpart -> raw_subtype), msg, &content_subtype_len);

		mpart -> type    = zm_compare_list(content_type_str, content_type_len, zm_types_str, zm_types);
		mpart -> subtype = OTHER_SUBTYPE;
		switch (mpart -> type)
		{
			case MULTIPART:
				{
					mpart -> u.multi.boundary.start = mpart -> u.multi.boundary.end = 0;

					if (content_type != NULL)
					{
						ZM_KEY_VAL  * kv = content_type -> value.pairs;
						while (kv != NULL)
						{
							uint64_t  key_len = 0;
							const unsigned char * key = zm_block_get_data(&(kv -> key), msg, &key_len);

							if (zm_compare((const char *)key, key_len, "boundary") == 0)
							{
								zm_block_copy(&(mpart -> u.multi.boundary), &(kv -> val));
								break;
							}

							kv = kv -> next;
						}
					}

					/* No boundary? - Fatal error */
					if (mpart -> u.multi.boundary.start == mpart -> u.multi.boundary.end) { return ZM_FAILED; }

					mpart -> subtype = zm_compare_list(content_subtype_str, content_subtype_len, zm_multipart_subtypes_str, zm_multipart_subtypes);
				}
				break;

			case MESSAGE:
				mpart -> subtype = zm_compare_list(content_subtype_str, content_subtype_len, zm_message_subtypes_str, zm_message_subtypes);
				break;

			case TEXT:
				{
					ZM_KEY_VAL  * kv = content_type -> value.pairs;

					mpart -> u.text.charset = US_ASCII;
					while (kv != NULL)
					{
						uint64_t  key_len = 0;
						const unsigned char * key = zm_block_get_data(&(kv -> key), msg, &key_len);

						if (zm_compare((const char *)key, key_len, "charset") == 0)
						{
							uint64_t       charset_str_len = 0;
							const char  *  charset_str     = (const char *)zm_block_get_data(&(kv -> val), msg, &charset_str_len);

							mpart -> u.text.charset = OTHER_CHARSET;

							mpart -> u.text.charset = zm_compare_list(charset_str, charset_str_len, zm_charsets_str, zm_charsets);
							/* Charset no found, search in aliases database */
							if (mpart -> u.text.charset == OTHER_CHARSET)
							{
								if (charset_aliases != NULL)
								{
									const char * charset_name = NULL;
									if (zm_parse_alias(charset_aliases, charset_str, charset_str_len, &charset_name) == 0)
									{
										mpart -> u.text.charset = zm_compare_list(charset_name, strlen(charset_name), zm_charsets_str, zm_charsets);
									}
								}
							}

							zm_block_copy(&(mpart -> u.text.raw_charset), &(kv -> val));

							break;
						}

						kv = kv -> next;
					}

					mpart -> subtype = zm_compare_list(content_subtype_str, content_subtype_len, zm_basic_subtypes_str, zm_basic_subtypes);
				}
				break;

			case BINARY_PART:
			case IMAGE:
			case AUDIO:
			case VIDEO:
			case APPLICATION:
			case OTHER_TYPE:
				mpart -> is_basic = 1;
				mpart -> subtype = zm_compare_list(content_subtype_str, content_subtype_len, zm_basic_subtypes_str, zm_basic_subtypes);
				break;
		}

		mpart -> content_type_val = &(content_type -> value);
	}

	tmp_len = 0;
	tmp     = (const char *)zm_block_get_data(&(mpart -> raw_content_transfer_encoding), msg, &tmp_len);
	if (tmp_len != 0) { mpart -> content_transfer_encoding = zm_compare_list(tmp, tmp_len, zm_cte_str, zm_cte); }
	else
	{
		/*
		 * FIXME: get REAL encoding, see RFC 1341
		 *
		 * The values "8bit", "7bit", and "binary" all imply that NO encoding has been performed.
		 * "7bit" means that the data is all represented as short lines of US-ASCII data.
		 * "8bit" means that the lines are short, but there may be non-ASCII characters (octets with the high-order bit set).
		 * "Binary" means that not only may non-ASCII characters be present, but also that the lines are not necessarily short enough for SMTP transport.
		 */
		mpart -> content_transfer_encoding = BINARY;
	}
	if (content_disposition != NULL)
	{
		tmp_len = 0;
		tmp     = (const char *)zm_block_get_data(&(mpart -> raw_content_disposition.raw), msg, &tmp_len);
		mpart -> content_disposition = zm_compare_list(tmp, tmp_len, zm_content_dispositions_str, zm_content_dispositions);
	}

	/* Parse multipart message */
	if (mpart -> type == MULTIPART) { return zm_parse_multipart(msg, it_start, it_part_end, it_end, boundary, parent, &mpart, line, alloc, charset_aliases, ignore_errors); }

	/* Parse message */
	if (mpart -> type == MESSAGE) { return zm_parse_message(msg, it_start, it_part_end, it_end, boundary, parent, &mpart, line, alloc, charset_aliases, ignore_errors); }

/* All other types (text/ basic/ ) */
return zm_parse_text_or_basic(msg, it_start, it_part_end, it_end, boundary, part, line);
}

/*
 * Parse MIME message
 */
int32_t zm_parse_mime(const void     * msg,
                      uint64_t         msg_len,
                      ZM_ALLOCATOR   * alloc,
                      int32_t          ignore_errors,
                      ZM_ALIASES     * charset_aliases,
                      ZM_PART       ** part)
{
	uint64_t     line = 0;

	ZM_ITERATOR  it_start;
	ZM_ITERATOR  it_part_end;
	ZM_ITERATOR  it_end;
	ZM_BLOCK     boundary;

	boundary.start = boundary.end = 0;

	zm_iterator_create(&it_start, msg, 0);
	zm_iterator_create(&it_part_end, msg, 0);
	zm_iterator_create(&it_end, msg, msg_len);

return zm_do_parse_mime(msg, &it_start, &it_part_end, &it_end, &boundary, NULL, part, &line, alloc, charset_aliases, ignore_errors) == ZM_END ? 0 : -1;
}
/* End. */
