/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_address_parser.h"
#include "zm_charset_detector.h"
#include "zm_data_buffer.h"
#include "zm_iconv.h"
#include "zm_mem.h"
#include "zm_mime_converter.h"
#include "zm_part_handler.h"
#include "zm_rbtree_cmp.h"
#include "zm_rbtree_fn.h"
#include "zm_struct.h"
#include "zm_time_parser.h"

#include <stdio.h>
#include <string.h>
#include <strings.h>

#ifndef ZM_BUF_NAME_LEN
#define ZM_BUF_NAME_LEN 127
#endif /* ZM_BUF_NAME_LEN */

#ifndef ZM_CHARSET_BUF_LEN
#define ZM_CHARSET_BUF_LEN 127
#endif /* ZM_CHARSET_BUF_LEN */

/**
 * @struct _zm_part_handler zm_part_handler.h <zm_part_handler.h>
 * @brief MIME part handler
 */
typedef struct _zm_part_handler_handler_rec    ZM_PART_HANDLER_HANDLER_REC;
struct ZM_API _zm_part_handler_handler_rec
{
	/** Opaque data for part handler */
	void                         * opaque;
	/** Start of MIME part           */
	zm_start_part_fn               start_part;
	/** Add MIME part data           */
	zm_add_part_data_fn            add_part_data;
	/** End of MIME part             */
	zm_end_part_fn                 end_part;

	/** Next record                  */
	ZM_PART_HANDLER_HANDLER_REC  * next;
};

/*
 * Fix header (remove CR?LF[\t ]
 */
static uint64_t zm_fix_header(void      * data,
                              uint64_t    data_len)
{
	unsigned char  * src = (unsigned char *)data;
	unsigned char  * dst = (unsigned char *)data;
	unsigned char  * end = (unsigned char *)data + data_len;

	while (src != end)
	{
		const unsigned char ch = *src;
		if (ch == '\r' || ch == '\n') { ;; }
		else if (ch <= ' ')
		{
			*dst++ = ' ';
		}
		else
		{
			*dst++ = ch;
		}
		++src;
	}
return dst - (unsigned char *)data;
}

/*
 * Create part name
 */
static int32_t zm_part_name_create_rec(ZM_PART         * part,
                                       ZM_DATA_BUFFER  * part_name)
{
	int32_t  rc = 0;
	char     name[ZM_BUF_NAME_LEN + 1];

	if (part == NULL) { return 0; }

	rc = zm_part_name_create_rec(part -> parent, part_name);
	if (rc != 0) { return rc; }

	if (part -> parent != NULL && !part -> is_text)
	{
		const int32_t name_len = sprintf(name, "%u.", part -> part_num);
		rc = zm_data_buffer_append_data(part_name, name, name_len) == -1 ? -1 : 0;
	}

return rc;
}

/*
 * Create part name
 */
static int32_t zm_part_name_create(ZM_PART         * part,
                                   ZM_DATA_BUFFER  * part_name)
{
	int32_t  rc = 0;
	char     name[ZM_BUF_NAME_LEN + 1];

	if (part == NULL) { return 0; }

	rc = zm_part_name_create_rec(part -> parent, part_name);
	if (rc != 0) { return rc; }

	if (part -> is_text)
	{
		rc = zm_data_buffer_append_data(part_name, "TEXT", sizeof("TEXT") - 1);
	}
	else
	{
		const int32_t name_len = sprintf(name, "%u", part -> part_num);
		rc = zm_data_buffer_append_data(part_name, name, name_len) == -1 ? -1 : 0;
	}

return rc;
}

/*
 * Deallocate charset map
 */
static void zm_part_handler_destroy_map_type(void  * opaque,
                                             void  * key,
                                             void  * val)
{
	ZM_RB_TREE * subtype_map = (ZM_RB_TREE *)val;

	zm_rb_destroy(subtype_map);
}

/*
 * Deallocate charset map
 */
static void zm_part_handler_destroy_map_subtype(void  * opaque,
                                                void  * key,
                                                void  * val)
{
	ZM_ALLOCATOR                * alloc       = (ZM_ALLOCATOR *)opaque;
	ZM_PART_HANDLER_HANDLER_REC * subtype_rec = (ZM_PART_HANDLER_HANDLER_REC *)val;

	while (subtype_rec != NULL)
	{
		ZM_PART_HANDLER_HANDLER_REC * tmp = subtype_rec -> next;
		zm_deallocate(alloc, subtype_rec);

		subtype_rec = tmp;
	}
}

/*
 * Create MIME part handler
 */
ZM_PART_HANDLER * zm_part_handler_create(ZM_PART_HANDLER    * part_handler,
                                         ZM_ICONV           * iconv,
                                         ZM_ABBREV          * abbrev,
                                         ZM_ADDRESS_PARSER  * address_parser,
                                         ZM_DATA_BUFFER     * buffer)
{
	uint32_t        flags = 0x00;
	ZM_ALLOCATOR  * alloc = zm_data_buffer_get_allocator(buffer);

	if (part_handler == NULL)
	{
		part_handler = (ZM_PART_HANDLER *)zm_allocate(alloc, sizeof(ZM_PART_HANDLER));
		flags = 0x01;
	}

	if (part_handler == NULL) { return NULL; }

	if (zm_envelope_create(&(part_handler -> envelope),            buffer) != NULL) { flags |= 0x10; }
	if (zm_envelope_create(&(part_handler -> decoded_envelope),    buffer) != NULL) { flags |= 0x20; }
	if (zm_data_buffer_clone(&(part_handler -> part_data),         buffer) != NULL) { flags |= 0x40; }
	if (zm_data_buffer_clone(&(part_handler -> part_data_recoded), buffer) != NULL) { flags |= 0x80; }

	if (zm_rb_create(&(part_handler -> handlers), alloc, zm_rb_void_cmp, zm_part_handler_destroy_map_type, alloc) != NULL) { flags |= 0x100; }

	part_handler -> iconv          = iconv;
	part_handler -> abbrev         = abbrev;
	part_handler -> address_parser = address_parser;
	part_handler -> alloc          = alloc;
	part_handler -> flags          = flags;

	if ((flags & 0x1F0) != 0x1F0)
	{
		zm_part_handler_destroy(part_handler);
		return NULL;
	}

return part_handler;
}

/*
 * Destroy MIME part handler
 */
void zm_part_handler_destroy(ZM_PART_HANDLER  * part_handler)
{
	ZM_ALLOCATOR  * alloc = part_handler -> alloc;

	if (part_handler -> flags & 0x10)  { zm_envelope_destroy(&(part_handler -> envelope));             }
	if (part_handler -> flags & 0x20)  { zm_envelope_destroy(&(part_handler -> decoded_envelope));     }
	if (part_handler -> flags & 0x40)  { zm_data_buffer_destroy(&(part_handler -> part_data));         }
	if (part_handler -> flags & 0x80)  { zm_data_buffer_destroy(&(part_handler -> part_data_recoded)); }
	if (part_handler -> flags & 0x100) { zm_rb_destroy(&(part_handler -> handlers)); }

	if (part_handler -> flags & 0x01)
	{
		zm_deallocate(alloc, part_handler);
	}
}

/*
 * Register MIME part handler
 */
int32_t zm_part_handler_register_handler(ZM_PART_HANDLER          * part_handler,
                                         enum ZM_CONTENT_TYPE       content_type,
                                         enum ZM_CONTENT_SUBTYPE    content_subtype,
                                         void                     * opaque,
                                         zm_start_part_fn           start_part_fn,
                                         zm_add_part_data_fn        add_part_data_fn,
                                         zm_end_part_fn             end_part_fn)
{
	ZM_ALLOCATOR                * alloc       = part_handler -> alloc;
	ZM_PART_HANDLER_HANDLER_REC * subtype_rec = NULL;
	ZM_PART_HANDLER_HANDLER_REC * tmp         = NULL;
	ZM_RB_TREE                  * subtype_map = NULL;

	subtype_map = zm_rb_search(&(part_handler -> handlers), (void *)content_type) -> val;
	if (subtype_map == ((void *)-1))
	{
		subtype_map = zm_rb_create(NULL, alloc, zm_rb_void_cmp, zm_part_handler_destroy_map_subtype, alloc);
		if (subtype_map == NULL) { return -1; }

		zm_rb_insert(&(part_handler -> handlers), (void *)content_type, subtype_map);
	}

	subtype_rec = zm_allocate(part_handler -> alloc, sizeof(ZM_PART_HANDLER_HANDLER_REC));
	if (subtype_rec == NULL) { return -1; }

	subtype_rec -> opaque        = opaque;
	subtype_rec -> start_part    = start_part_fn;
	subtype_rec -> add_part_data = add_part_data_fn;
	subtype_rec -> end_part      = end_part_fn;

	tmp = zm_rb_search(subtype_map, (void *)content_subtype) -> val;
	if (tmp == ((void *)-1))
	{
		zm_rb_insert(subtype_map, (void *)content_subtype, subtype_rec);
		return 0;
	}

	while (tmp -> next != NULL) { tmp = tmp -> next; }

	tmp -> next = subtype_rec;

return 0;
}

/*
 * Get chain of handlers for specified MIME part
 */
static ZM_PART_HANDLER_HANDLER_REC * zm_part_handler_get_handler(ZM_PART_HANDLER          * part_handler,
                                                                 enum ZM_CONTENT_TYPE       content_type,
                                                                 enum ZM_CONTENT_SUBTYPE    content_subtype)
{
	/* Type handlers */
	ZM_PART_HANDLER_HANDLER_REC * subtype_rec = NULL;
	ZM_RB_TREE                  * subtype_map = zm_rb_search(&(part_handler -> handlers), (void *)content_type) -> val;
	if (subtype_map == ((void *)-1) || subtype_map == NULL)
	{
		/* Check OTHER type handlers */
		subtype_map = zm_rb_search(&(part_handler -> handlers), (void *)OTHER_TYPE) -> val;
		if (subtype_map == ((void *)-1) || subtype_map == NULL) { return NULL; }
	}

	/* Subtype handlers */
	subtype_rec = zm_rb_search(subtype_map, (void *)content_subtype) -> val;
	if (subtype_rec == ((void *)-1) || subtype_rec == NULL)
	{
		/* Check OTHER subtype handlers */
		subtype_rec = zm_rb_search(subtype_map, (void *)OTHER_SUBTYPE) -> val;
		if (subtype_rec == ((void *)-1) || subtype_rec == NULL) { return NULL; }
	}

return subtype_rec;
}

/*
 * Decode email part
 */
static int32_t zm_part_handler_decode_email_part(ZM_PART_HANDLER  * part_handler,
                                                 ZM_DATA_BUFFER   * dst,
                                                 ZM_DATA_BUFFER   * src)
{
	const char * value     = zm_data_buffer_get(src);
	int32_t      value_len = zm_data_buffer_get_size(src);

return zm_decode_mime_encoded((const unsigned char *)value, value_len, dst, part_handler -> iconv) != NULL;
}

/*
 * Decode Email list
 */
static int32_t zm_part_handler_decode_email_list(ZM_PART_HANDLER    * part_handler,
                                                 const char         * value,
                                                 uint64_t             value_len,
                                                 ZM_EMAIL_ADDRESS  ** src_list,
                                                 ZM_EMAIL_ADDRESS  ** dst_list)
{
	int32_t rc = 0;
	ZM_EMAIL_ADDRESS  * src;
	ZM_EMAIL_ADDRESS  * dst;

	const unsigned char  * addr_src     = (const unsigned char *)value;
	uint64_t               addr_src_len = value_len;

	if (value_len == 0) { return 0; }

	rc = zm_address_parser_parse_address_list(part_handler -> address_parser, &addr_src, &addr_src_len, src_list);
	if (rc != 0 && rc != ZM_ADDRESS_EOL) { return -1; }

	src = *src_list;
	dst = *dst_list;
	while (src != NULL)
	{
		ZM_EMAIL_ADDRESS  * next_src = src -> next;

		ZM_EMAIL_ADDRESS  * email_address = zm_email_address_create(NULL, &(part_handler -> part_data));

		if (email_address == NULL) { return -1; }

		if (dst == NULL)
		{
			*dst_list = dst = email_address;
		}
		else
		{
			dst -> next = email_address;
			dst = email_address;
		}

		zm_part_handler_decode_email_part(part_handler, &(email_address -> group),    &(src -> group));
		zm_part_handler_decode_email_part(part_handler, &(email_address -> name),     &(src -> name));
		zm_part_handler_decode_email_part(part_handler, &(email_address -> route),    &(src -> route));
		zm_part_handler_decode_email_part(part_handler, &(email_address -> username), &(src -> username));
		zm_part_handler_decode_email_part(part_handler, &(email_address -> domain),   &(src -> domain));

		src = next_src;
	}

return 0;
}

/*
 * Decode envelope fields
 */
static int32_t zm_part_handler_extract_envelope(ZM_PART_HANDLER  * part_handler,
                                                const void       * data,
                                                ZM_PART          * part)
{
	ZM_FIELD_REC  * field = NULL;
	uint32_t        flags = 0;
	if (part == NULL) { return ZM_PART_ERROR; }

	field = zm_part_first_header(part);
	while (field != NULL)
	{
		int64_t       new_size  = 0;
		uint64_t      name_len  = 0;
		const char  * name      = zm_block_get_data(&(field -> name), data, &name_len);

		uint64_t      value_len = 0;
		const char  * value     = zm_block_get_data(&(field -> value.raw), data, &value_len);

		if (strncasecmp(name, "subject", name_len) == 0)
		{
			flags |= 0x01;
			zm_data_buffer_append_data(&(part_handler -> envelope.subject), value, value_len);
			new_size = zm_fix_header(zm_data_buffer_get(&(part_handler -> envelope.subject)), zm_data_buffer_get_size(&(part_handler -> envelope.subject)));
			zm_data_buffer_set_size(&(part_handler -> envelope.subject), new_size);

			zm_decode_mime_encoded((const unsigned char *)value, value_len, &(part_handler -> decoded_envelope.subject), part_handler -> iconv);
		}
		else if (strncasecmp(name, "from", name_len) == 0)
		{
			flags |= 0x02;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.from), &(part_handler -> decoded_envelope.from));
		}
		else if (strncasecmp(name, "sender", name_len) == 0)
		{
			flags |= 0x04;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.sender), &(part_handler -> decoded_envelope.sender));
		}
		else if (strncasecmp(name, "reply-to", name_len) == 0)
		{
			flags |= 0x08;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.reply_to), &(part_handler -> decoded_envelope.reply_to));
		}
		else if (strncasecmp(name, "to", name_len) == 0)
		{
			flags |= 0x10;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.to), &(part_handler -> decoded_envelope.to));
		}
		else if (strncasecmp(name, "cc", name_len) == 0)
		{
			flags |= 0x20;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.cc), &(part_handler -> decoded_envelope.cc));
		}
		else if (strncasecmp(name, "bcc", name_len) == 0)
		{
			flags |= 0x40;
			zm_part_handler_decode_email_list(part_handler, value, value_len, &(part_handler -> envelope.bcc), &(part_handler -> decoded_envelope.bcc));
		}
		else if (strncasecmp(name, "date", name_len) == 0)
		{
			flags |= 0x80;
			zm_data_buffer_assign_data(&(part_handler -> envelope.date),         value, value_len);
			zm_data_buffer_assign_data(&(part_handler -> decoded_envelope.date), value, value_len);

			zm_parse_datetime(value, value_len, &(part_handler -> envelope.date_ts), &(part_handler -> envelope.date_ts_tz), part_handler -> abbrev);
			part_handler -> decoded_envelope.date_ts    = part_handler -> envelope.date_ts;
			part_handler -> decoded_envelope.date_ts_tz = part_handler -> envelope.date_ts_tz;
		}
		else if (strncasecmp(name, "in-reply-to", name_len) == 0)
		{
			flags |= 0x100;
			zm_data_buffer_assign_data(&(part_handler -> envelope.in_reply_to),         value, value_len);
			zm_data_buffer_assign_data(&(part_handler -> decoded_envelope.in_reply_to), value, value_len);
		}
		else if (strncasecmp(name, "message-id", name_len) == 0)
		{
			flags |= 0x200;
			zm_data_buffer_assign_data(&(part_handler -> envelope.message_id),         value, value_len);
			zm_data_buffer_assign_data(&(part_handler -> decoded_envelope.message_id), value, value_len);
		}

		/* Nothing to do? */
		if (flags == 0x3FF) { break; }

		field = zm_part_header_next(field);
	}

return ZM_PART_OK;
}

/*
 * Decode part data
 */
static int32_t zm_part_handler_extract_part(ZM_PART_HANDLER  * part_handler,
                                            const void       * data,
                                            ZM_PART          * part,
                                            ZM_DATA_BUFFER   * part_data)
{
	switch(part -> content_transfer_encoding)
	{
		case QUOTED_PRINTABLE:
			return zm_decode_quoted_printable(&(part -> data.raw), data, part_data) == NULL ? ZM_PART_ERROR : ZM_PART_OK;
		case BASE64:
			return zm_decode_base64(&(part -> data.raw), data, zm_base64_rev, part_data) == NULL ? ZM_PART_ERROR : ZM_PART_OK;
		case UU_ENCODED:
			return zm_decode_uu_encoded(&(part -> data.raw), data, part_data) == NULL ? ZM_PART_ERROR : ZM_PART_OK;
		default:
			return ZM_PART_RAW_DATA;
	}

return ZM_PART_RAW_DATA;
}

/*
 * Decode MIME-encoded header
 */
int32_t zm_part_handler_decode_mime_encoded(ZM_PART_HANDLER  * part_handler,
                                            const void       * data,
                                            const ZM_BLOCK   * block,
                                            ZM_DATA_BUFFER   * dst)
{
	uint64_t     value_len = 0;
	const void * value     = zm_block_get_data(block, data, &value_len);

return zm_decode_mime_encoded((const unsigned char *)value, value_len, dst, part_handler -> iconv) != NULL;
}

/*
 * Extract data
 */
int32_t zm_part_handler_extract_data(ZM_PART_HANDLER  * part_handler,
                                     const void       * data,
                                     const uint64_t     data_len,
                                     ZM_PART          * part,
                                     const uint32_t     flags)
{
	int32_t           extracted = 0;
	int32_t           rc        = ZM_PART_OK;
	ZM_DATA_BUFFER    part_name;
	const void      * part_data = NULL;
	uint64_t          part_data_len = 0;

	ZM_PART_HANDLER_HANDLER_REC  * handler_rec       = NULL;
	ZM_PART_HANDLER_HANDLER_REC  * handler_rec_first = NULL;

	if (part == NULL) { return ZM_PART_ERROR; }

	handler_rec_first = zm_part_handler_get_handler(part_handler, part -> type, part -> subtype);
	if (handler_rec_first == NULL) { return ZM_PART_ERROR; }

	if (zm_data_buffer_clone(&part_name, &(part_handler -> part_data)) == NULL) { return ZM_PART_ERROR; }

	if (zm_part_name_create(part, &part_name) == -1)
	{
		rc = ZM_PART_ERROR;
		goto EXTRACT_DATA_END;
	}

	zm_data_buffer_reset(&(part_handler -> part_data));

	if (flags & ZM_RECODE_TEXT_PART) { zm_data_buffer_reset(&(part_handler -> part_data_recoded)); }

	handler_rec = handler_rec_first;
	while (handler_rec != NULL)
	{
		if (handler_rec -> start_part != NULL)
		{
			if (flags & ZM_EXTRACT_ENVELOPE)
			{
				if (!(extracted & 0x01))
				{
					rc = zm_part_handler_extract_envelope(part_handler, data, part);
					if (rc == ZM_PART_ERROR) { goto EXTRACT_DATA_END; }

					extracted |= 0x01;
				}

				rc = handler_rec -> start_part(handler_rec -> opaque, data, part, &part_name, &(part_handler -> envelope), &(part_handler -> decoded_envelope));
			}
			else
			{
				rc = handler_rec -> start_part(handler_rec -> opaque, data, part, &part_name, NULL, NULL);
			}

			if (rc == ZM_PART_ERROR || rc == ZM_PART_BREAK) { goto EXTRACT_DATA_END; }

			/* Extract data, if need */
			if (rc == ZM_PART_NEED_DATA)
			{
				if (!(extracted & 0x02))
				{
					rc = zm_part_handler_extract_part(part_handler, data, part, &(part_handler -> part_data));
					if (rc == ZM_PART_ERROR) { goto EXTRACT_DATA_END; }

					if (rc == ZM_PART_OK)
					{
						part_data     = zm_data_buffer_get(&(part_handler -> part_data));
						part_data_len = zm_data_buffer_get_size(&(part_handler -> part_data));
					}
					else /* if (rc == ZM_PART_RAW_DATA) */
					{
						part_data = zm_block_get_data(&(part -> data.raw), data, &part_data_len);
					}

					if ((flags & ZM_RECODE_TEXT_PART) && part -> type == TEXT)
					{
						char charset_buf[ZM_CHARSET_BUF_LEN + 1];

						uint32_t      errors      = 0;
						uint64_t      charset_len = 0;
						const char  * charset     = zm_get_text_part_charset(&(part -> u.text), data, &charset_len);

						if (charset_len < ZM_CHARSET_BUF_LEN)
						{
							memcpy(charset_buf, charset, charset_len);
							charset_buf[charset_len] = '\0';

							rc = zm_iconv_convert_data_buffer(part_handler -> iconv,
							                                   part_data, part_data_len,
							                                   &(part_handler -> part_data_recoded),
							                                   charset_buf,
							                                   "UTF-8",
							                                   &errors,
							                                   1);

							if (errors > part_data_len / 4)
							{
								const char  * detected_charset = "US-ASCII";
								enum ZM_CHARSET charset = zm_detect_charset((const unsigned char *)part_data, part_data_len);
								switch(charset)
								{
									case UTF_8:        detected_charset = "UTF-8"; break;
									case WINDOWS_1251: detected_charset = "CP1251"; break;
									case KOI_8R:       detected_charset = "KOI"; break;
									default:
										detected_charset = "US-ASCII";
								}

								rc = zm_iconv_convert_data_buffer(part_handler -> iconv,
								                                   part_data, part_data_len,
								                                   &(part_handler -> part_data_recoded),
								                                   detected_charset,
								                                   "UTF-8",
								                                   &errors,
								                                   1);
							}

							if (rc == 0)
							{
								part_data     = zm_data_buffer_get(&(part_handler -> part_data_recoded));
								part_data_len = zm_data_buffer_get_size(&(part_handler -> part_data_recoded));
							}
						}
					}
					extracted |= 0x02;
				}

				rc = handler_rec -> add_part_data(handler_rec -> opaque, data, part, &part_name, part_data, part_data_len);
				if (rc == ZM_PART_ERROR || rc == ZM_PART_BREAK) { goto EXTRACT_DATA_END; }
			}
			/* Error */
			else if (rc != ZM_PART_OK && rc != ZM_PART_NEED_SUBPARTS) { goto EXTRACT_DATA_END; }
		}

		zm_envelope_reset(&(part_handler -> envelope));
		zm_envelope_reset(&(part_handler -> decoded_envelope));

		handler_rec = handler_rec -> next;
	}

	/* Handle MESSAGE subpart */
	if (part -> type == MESSAGE)
	{
		rc = zm_part_handler_extract_data(part_handler, data, data_len, zm_message_subpart(part), flags);
		if (rc == ZM_PART_ERROR || rc == ZM_PART_BREAK) { goto EXTRACT_DATA_END; }
	}
	/* Handle MULTIPART subparts */
	else if (part -> type == MULTIPART)
	{
		ZM_PART  * mpart = zm_multipart_first_subpart(part);
		while (mpart != NULL)
		{
			rc = zm_part_handler_extract_data(part_handler, data, data_len, mpart, flags);
			if (rc == ZM_PART_ERROR || rc == ZM_PART_BREAK) { goto EXTRACT_DATA_END; }

			mpart = zm_next_part(mpart);
		}
	}

	handler_rec = handler_rec_first;
	while (handler_rec != NULL)
	{
		if (handler_rec -> end_part != NULL)
		{
			rc = handler_rec -> end_part(handler_rec -> opaque, data, part, &part_name);
			if (rc == ZM_PART_ERROR || rc == ZM_PART_BREAK) { goto EXTRACT_DATA_END; }
		}

		handler_rec = handler_rec -> next;
	}

EXTRACT_DATA_END:
	zm_data_buffer_destroy(&part_name);

return rc;
}
/* End. */

