/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include "zm_mem.h"
#include "zm_rbtree.h"

/*
 * Perform a left rotation starting at node.
 */
static void zm_rb_rotate_left(ZM_RB_TREE  * tree,
                              ZM_RB_NODE  * node)
{
	ZM_RB_NODE  * child = node -> right;
	ZM_RB_NODE  * nil   = &(tree -> nil);

	node -> right = child -> left;

	if (child -> left != nil) { child -> left -> parent = node; }

	child -> parent = node -> parent;

	if(node == node -> parent -> left) { node -> parent -> left  = child; }
	else                               { node -> parent -> right = child; }

	child -> left   = node;
	node -> parent = child;
}

/*
 * Perform a right rotation starting at node.
 */
static void zm_rb_rotate_right(ZM_RB_TREE  * tree,
                               ZM_RB_NODE  * node)
{
	ZM_RB_NODE  * child = node -> left;
	ZM_RB_NODE  * nil   = &(tree -> nil);

	node -> left = child -> right;

	if (child -> right != nil) { child -> right -> parent = node; }

	child -> parent = node -> parent;

	if (node == node -> parent -> left) { node -> parent -> left  = child; }
	else                                { node -> parent -> right = child; }

	child -> right = node;
	node -> parent = child;
}

/*
 * Recursive dump tree
 */
static int32_t zm_rb_dump_rec(ZM_RB_TREE          * tree,
                              ZM_RB_NODE          * node,
                              void                  * opaque,
                              zm_rb_tree_dump_fn    dump)
{
	ZM_RB_NODE * nil  = &(tree -> nil);
	int32_t rc = 0;

	if (node == nil) { return 0; }

	rc = zm_rb_dump_rec(tree, node -> left, opaque, dump);
	if (rc == -1) { return -1; }

	if (dump(opaque, node -> pair.key, node -> pair.val) != 0) { return -1; }

return zm_rb_dump_rec(tree, node -> right, opaque, dump);
}

/*
 * Destroy tree
 */
static void zm_rb_destroy_rec(ZM_RB_TREE  * tree,
                              ZM_RB_NODE  * node)
{
	ZM_ALLOCATOR  * alloc = tree -> alloc;
	ZM_RB_NODE    * nil   = &(tree -> nil);

	if (node != nil)
	{
		zm_rb_destroy_rec(tree, node -> left);
		zm_rb_destroy_rec(tree, node -> right);

		tree -> destroy(tree -> opaque, node -> pair.key, node -> pair.val);

		zm_deallocate(alloc, node);
	}
}

/*
 * Get successor
 */
static ZM_RB_NODE * zm_rb_tree_successor(ZM_RB_TREE  * tree,
                                         ZM_RB_NODE  * node)
{
	ZM_RB_NODE  * root = &(tree -> root);
	ZM_RB_NODE  * nil  = &(tree -> nil);
	ZM_RB_NODE  * y    = node -> right;

	if (y != nil)
	{
		while(y -> left != nil) { y = y -> left; }
		return y;
	}

	y = node -> parent;
	while(node == y -> right)
	{
		node = y;
		y = y -> parent;
	}

	if (y == root) { return nil; }

return y;
}

/*
 * Repair the tree after a node has been deleted by rotating and repainting
 * colors to restore the 4 properties inherent in red-black trees.
 */
static void zm_rb_repair(ZM_RB_TREE  * tree,
                         ZM_RB_NODE  * node)
{
	ZM_RB_NODE  * sibling = NULL;

	while (node -> red == 0)
	{
		if (node == node -> parent -> left)
		{
			sibling = node -> parent -> right;
			if (sibling -> red == 1)
			{
				sibling-> red = 0;
				node -> parent -> red = 1;
				zm_rb_rotate_left(tree, node -> parent);
				sibling = node -> parent -> right;
			}

			if (sibling -> right -> red == 0 && sibling -> left -> red == 0)
			{
				sibling-> red = 1;
				node = node -> parent;
			}
			else
			{
				if (sibling -> right -> red == 0)
				{
					sibling -> left -> red = 0;
					sibling -> red = 1;
					zm_rb_rotate_right(tree, sibling);
					sibling = node -> parent -> right;
				}
				sibling -> red = node -> parent -> red;
				node -> parent -> red = 0;
				sibling -> right -> red = 0;
				zm_rb_rotate_left(tree, node -> parent);
				break;
			}
		}
		else /* if (node == node -> parent -> right) */
		{
			sibling = node -> parent -> left;
			if (sibling -> red == 1)
			{
				sibling -> red = 0;
				node -> parent -> red = 1;
				zm_rb_rotate_right(tree, node -> parent);
				sibling = node -> parent -> left;
			}

			if (sibling -> right -> red == 0 && sibling -> left -> red == 0)
			{
				sibling -> red = 1;
				node = node -> parent;
			}
			else
			{
				if (sibling -> left -> red == 0)
				{
					sibling -> right -> red = 0;
					sibling-> red = 1;
					zm_rb_rotate_left(tree, sibling);
					sibling = node -> parent -> left;
				}
				sibling -> red = node -> parent -> red;
				node -> parent -> red = 0;
				sibling -> left -> red = 0;
				zm_rb_rotate_right(tree, node -> parent);
				break;
			}
		}
	}
}

/*
 * Create tree
 */
ZM_RB_TREE * zm_rb_create(ZM_RB_TREE             * tree,
                          void                   * opaque,
                          zm_rb_tree_cmp_fn        cmp,
                          zm_rb_tree_destroy_fn    destroy,
                          ZM_ALLOCATOR           * alloc)
{

	ZM_RB_NODE * nil;
	uint32_t flags = 0;

	if (tree == NULL)
	{
		tree = (ZM_RB_TREE*)zm_allocate(alloc, sizeof(ZM_RB_TREE));
		flags = 1;
	}

	if (tree == NULL) { return NULL; }

	nil = &(tree -> nil);

	tree -> flags    = flags;
	tree -> size     = 0;
	tree -> opaque   = opaque;
	tree -> cmp      = cmp;
	tree -> destroy  = destroy;

	tree -> alloc    = alloc;

	tree -> nil.parent    = nil;
	tree -> nil.left      = nil;
	tree -> nil.right     = nil;
	tree -> nil.red       = 0;
	tree -> nil.pair.key  = (void*)(-1);
	tree -> nil.pair.val  = (void*)(-1);

	tree -> root.parent   = nil;
	tree -> root.left     = nil;
	tree -> root.right    = nil;
	tree -> root.pair.key = (void*)(-1);
	tree -> root.pair.val = (void*)(-1);
	tree -> root.red      = 0;

return tree;
}

/*
 * Destroy tree
 */
void zm_rb_destroy(ZM_RB_TREE  * tree)
{
	ZM_ALLOCATOR  * alloc = tree -> alloc;

	zm_rb_destroy_rec(tree, tree -> root.left);

	if (tree -> flags == 1) { zm_deallocate(alloc, tree); }
}

/*
 * Insert data pointer into a redblack tree.
 * Returns a NULL pointer on success.  If a node matching "data"
 * already exists, a pointer to the existant node is returned.
 */
ZM_RB_NODE * zm_rb_insert_node(ZM_RB_TREE  * tree,
                               void          * key,
                               void          * val)
{
	ZM_ALLOCATOR  * alloc  = tree -> alloc;
	ZM_RB_NODE    * nil    = &(tree -> nil);
	ZM_RB_NODE    * node   = tree -> root.left;
	ZM_RB_NODE    * parent = &(tree -> root);

	while (node != nil)
	{
		int32_t cmp = tree -> cmp(tree -> opaque, key, node -> pair.key);
		if (cmp == 0) { return node; }

		parent = node;

		node = cmp < 0 ? node -> left : node -> right;
	}

	node = (ZM_RB_NODE *)zm_allocate(alloc, sizeof(ZM_RB_NODE));
	if (node == NULL) { return (ZM_RB_NODE *)-1; }

	node -> pair.key = key;
	node -> pair.val = val;
	node -> left     = nil;
	node -> right    = nil;
	node -> parent   = parent;
	node -> red      = 1;

	if (parent == &(tree -> root) ||
	    tree -> cmp(tree -> opaque, key, parent -> pair.key) < 0
	   )
	{
		parent -> left = node;
	}
	else
	{
		parent -> right = node;
	}

	while (node -> parent -> red)
	{
		ZM_RB_NODE *uncle;
		if (node -> parent == node -> parent -> parent -> left)
		{
			uncle = node -> parent -> parent -> right;
			if (uncle -> red == 1)
			{
				node -> parent -> red = 0;
				uncle -> red = 0;
				node -> parent -> parent -> red = 1;
				node = node -> parent -> parent;
			}
			else /* if (uncle -> red == 0) */
			{
				if (node == node -> parent -> right)
				{
					node = node -> parent;
					zm_rb_rotate_left(tree, node);
				}

				node -> parent -> red = 0;
				node -> parent -> parent -> red = 1;

				zm_rb_rotate_right(tree, node -> parent -> parent);
			}
		}
		else /* if (node -> parent == node -> parent -> parent -> right) */
		{
			uncle = node -> parent -> parent -> left;
			if (uncle -> red == 1)
			{
				node -> parent -> red = 0;
				uncle -> red = 0;
				node -> parent -> parent -> red = 1;
				node = node -> parent -> parent;
			}
			else /* if (uncle -> red == 0) */
			{
				if (node == node -> parent -> left)
				{
					node = node -> parent;
					zm_rb_rotate_right(tree, node);
				}

				node -> parent -> red = 0;
				node -> parent -> parent -> red = 1;

				zm_rb_rotate_left(tree, node -> parent -> parent);
			}
		}
	}

	tree -> root.left -> red = 0;
	tree -> size += 1;

return NULL;
}

/*
 * Look for a node matching key in tree.
 * Returns a pointer to the node if found, else NULL.
 */
ZM_RB_NODE * zm_rb_search_node(ZM_RB_TREE  * tree,
                               const void    * key)
{
	ZM_RB_NODE  * nil  = &(tree -> nil);
	ZM_RB_NODE  * node = tree -> root.left;

	while (node != nil)
	{
		int32_t cmp = tree -> cmp(tree -> opaque, key, node -> pair.key);
		if (cmp == 0) { return node; }

		node = cmp < 0 ? node -> left : node -> right;
	}

return NULL;
}

/*
 * Delete node 'z' from the tree and return its data pointer.
 */
void zm_rb_delete_node(ZM_RB_TREE  * tree,
                       ZM_RB_NODE  * z)
{
	ZM_ALLOCATOR  * alloc = tree -> alloc;
	ZM_RB_NODE    * nil   = &(tree -> nil);
	ZM_RB_NODE    * x;
	ZM_RB_NODE    * y;

	if (z -> left == nil || z -> right == nil)
	{
		y = z;
	}
	else
	{
		y = zm_rb_tree_successor(tree, z);
	}

	x = (y -> left == nil) ? y -> right : y -> left;

	x -> parent = y -> parent;
	if (x -> parent == &(tree -> root))
	{
		tree -> root.left = x;
	}
	else
	{
		if (y == y -> parent -> left) { y -> parent -> left  = x; }
		else                          { y -> parent -> right = x; }
	}

	if (y -> red == 0) { zm_rb_repair(tree, x); }

	if (y != z)
	{
		y -> left   = z -> left;
		y -> right  = z -> right;
		y -> parent = z -> parent;
		y -> red  = z -> red;
		z -> left -> parent = z -> right -> parent = y;

		if (z == z -> parent -> left) { z -> parent -> left  = y; }
		else                          { z -> parent -> right = y; }
	}

	tree -> destroy(tree -> opaque, z -> pair.key, z -> pair.val);

	zm_deallocate(alloc, z);

	tree -> size -= 1;
}

/*
 * Insert data pointer into a redblack tree.
 */
int32_t zm_rb_insert(ZM_RB_TREE  * tree,
                     void        * key,
                     void        * val)
{
	ZM_RB_NODE * node = zm_rb_insert_node(tree, key, val);
	if (node == NULL) { return 0; }

return -1;
}

/*
 * Look for a value matching key in tree. Returns a pointer to the node if found, else NULL.
 */
ZM_RB_PAIR * zm_rb_search(ZM_RB_TREE  * tree,
                          const void  * key)
{
	ZM_RB_NODE * node = zm_rb_search_node(tree, key);
	if (node == NULL) { return &(tree -> nil.pair); }

return &(node -> pair);
}

/*
 * Delete node from tree
 */
int32_t zm_rb_delete(ZM_RB_TREE  * tree,
                     const void  * key)
{
	ZM_RB_NODE * node = zm_rb_search_node(tree, key);
	if (node == NULL) { return -1; }

	zm_rb_delete_node(tree, node);

return 0;
}

/*
 * Clear tree, destruct all elements
 */
void zm_rb_clear(ZM_RB_TREE  * tree)
{
	zm_rb_destroy_rec(tree, tree -> root.left);
}

/*
 * Dump tree
 */
int32_t zm_rb_dump(ZM_RB_TREE          * tree,
                   void                * opaque,
                   zm_rb_tree_dump_fn    dump)
{
	return zm_rb_dump_rec(tree, tree -> root.left, opaque, dump);
}
/* End */
