/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <sysexits.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>

#include "zm_address_parser.h"
#include "zm_aliases.h"
#include "zm_data_buffer_impl.h"
#include "zm_block.h"
#include "zm_iconv.h"
#include "zm_part_handler.h"
#include "zm_mem.h"
#include "zm_parser.h"
#include "zm_struct.h"

static int32_t level = 0;

/*
 * Print part offset
 */
static void print_part_offset(void)
{
	int32_t pos = 0;

	for (; pos < level; ++pos) { fprintf(stdout, "    "); }
}

/*
 * Print mail header
 */
static void print_header(const char            * name,
                         const ZM_DATA_BUFFER  * value)
{
	const char     * dst_str = zm_data_buffer_get(value);
	const int32_t    dst_len = zm_data_buffer_get_size(value);

	if (dst_len)
	{
		print_part_offset();
		fprintf(stdout, "%s: `%.*s`\n", name, dst_len, dst_str);
	}
}

/*
 * Print mail header
 */
static int32_t print_email_part(const ZM_DATA_BUFFER  * value)
{
	const char     * dst_str = zm_data_buffer_get(value);
	const int32_t    dst_len = zm_data_buffer_get_size(value);

	if (dst_len == 0) { return 0; }

	fprintf(stdout, "`%.*s` ", dst_len, dst_str);

return 1;
}

/*
 * Print address chain
 */
static void print_address(const char        * name,
                          ZM_EMAIL_ADDRESS  * email_address)
{
	if (email_address == NULL) { return; }
	print_part_offset();
	fprintf(stdout, "%s: ", name);
	while (email_address != NULL)
	{
		fprintf(stdout, "(");

		print_email_part(&(email_address -> route));
		print_email_part(&(email_address -> group));
		print_email_part(&(email_address -> name));
		if (print_email_part(&(email_address -> username)))
		{
			fprintf(stdout, "@ ");
		}
		print_email_part(&(email_address -> domain));

		fprintf(stdout, ") ");

		email_address = zm_email_address_next(email_address);
	}
	fprintf(stdout, "\n");
}

/*
 * Print envelope
 */
static void print_envelope(const ZM_ENVELOPE  * envelope)
{
	print_header("Date     ", &(envelope -> date));
	if (envelope -> date_ts != -1)
	{
		print_part_offset();
		fprintf(stdout, "Unixtime : %lld, %lld\n", (long long int)(envelope -> date_ts), (long long int)(envelope -> date_ts_tz));
	}

	print_header("Subject  ", &(envelope -> subject));

	print_address("From     ", envelope -> from);
	print_address("To       ", envelope -> to);
	print_address("Cc       ", envelope -> cc);
	print_address("Bcc      ", envelope -> bcc);
	print_address("Sender   ", envelope -> sender);
	print_address("Reply-To ", envelope -> reply_to);

	print_header("MessageId", &(envelope -> message_id));
}

/*
 * Start of MIME part
 */
static int32_t zm_start_part(void                  * opaque,
                             const void            * data,
                             const ZM_PART         * part,
                             const ZM_DATA_BUFFER  * part_name,
                             const ZM_ENVELOPE     * envelope,
                             const ZM_ENVELOPE     * decoded_envelope)
{
	uint64_t      type_len;
	const char  * type;

	uint64_t      subtype_len;
	const char  * subtype;

	++level;

	print_part_offset();
	fprintf(stdout, "Start Part: %.*s ", (int32_t)zm_data_buffer_get_size(part_name), zm_data_buffer_get(part_name));

	type    = zm_block_get_data(&(part -> raw_type), data, &type_len);
	subtype = zm_block_get_data(&(part -> raw_subtype), data, &subtype_len);
	fprintf(stdout, "%.*s/%.*s\n", (int32_t)type_len, type, (int32_t)subtype_len, subtype);

	print_envelope(envelope);
	fprintf(stdout, "\n");
	print_envelope(decoded_envelope);

	print_part_offset();
	fprintf(stdout, "Length/Lines: %u/%d\n", (int32_t)(part -> data.raw.end - part -> data.raw.start), (int32_t)(part -> data.lines));

	if (part -> type == MULTIPART) { return ZM_PART_NEED_SUBPARTS; }
return ZM_PART_NEED_DATA;
}

/*
 * Add MIME part data
 */
static int32_t zm_add_part_data(void                  * opaque,
                                const void            * data,
                                const ZM_PART         * part,
                                const ZM_DATA_BUFFER  * part_name,
                                const void            * part_data,
                                const uint64_t          part_data_len)
{
	print_part_offset();
	fprintf(stdout, "Part Data: %.*s; length: %u\n", (int32_t)zm_data_buffer_get_size(part_name), zm_data_buffer_get(part_name), (uint32_t)part_data_len);

return ZM_PART_OK;
}

/*
 * End of MIME part
 */
static int32_t zm_end_part(void                  * opaque,
                           const void            * data,
                           const ZM_PART         * part,
                           const ZM_DATA_BUFFER  * part_name)
{

	print_part_offset();
	fprintf(stdout, "End Part: %.*s\n", (int32_t)zm_data_buffer_get_size(part_name), zm_data_buffer_get(part_name));
	--level;

return ZM_PART_OK;
}

int main(int argc, char ** argv)
{
	ZM_ALIASES   * aliases = NULL;
	int32_t        fh      = 0;
	void         * msg     = NULL;
	uint64_t       msg_len = 0;

	ZM_PART      * part = NULL;
	struct stat    st;

	if (argc != 2 && argc != 3) { fprintf(stderr, "usage: %s message.mime [aliases.txt]\n", argv[0]); return EX_USAGE; }

	fh = open(argv[1], O_NONBLOCK, O_RDONLY);
	if (fh == -1)
	{
		fprintf(stderr, "Can't open file `%s`\n", argv[1]);
		return EX_SOFTWARE;
	}

	if (fstat(fh, &st) == -1)
	{
		fprintf(stderr, "Can't stat file `%s`\n", argv[1]);
		close(fh);
		return EX_SOFTWARE;
	}

	msg_len = st.st_size;

	msg = mmap(NULL, msg_len, PROT_READ, MAP_PRIVATE, fh, 0);
	if (msg == MAP_FAILED)
	{
		fprintf(stderr, "Can't map file `%s` into memory\n", argv[1]);
		close(fh);
		return EX_SOFTWARE;
	}

	if (argc == 3)
	{
		if (zm_load_aliases(argv[2], &zm_default_allocator, &aliases) == -1) { return EX_SOFTWARE; }
	}

	zm_parse_mime(msg, msg_len, &zm_default_allocator, 1, aliases, &part);

	if (part != NULL)
	{
		ZM_DATA_BUFFER     buffer;
		ZM_PART_HANDLER    part_handler;
		ZM_ICONV           iconv;
		ZM_ADDRESS_PARSER  address_parser;
		ZM_ABBREV        * abbrev = NULL;

		zm_data_buffer_create_impl(&buffer, &zm_default_allocator);
		zm_iconv_create(&iconv, &zm_default_allocator);
		zm_address_parser_create(&address_parser, &iconv, &buffer);

		zm_part_handler_create(&part_handler, &iconv, abbrev, &address_parser, &buffer);
		/* zm_part_handler_register_handler(&part_handler, OTHER_TYPE, OTHER_SUBTYPE, NULL, zm_start_part, zm_add_part_data, zm_end_part); */
		zm_part_handler_register_handler(&part_handler, OTHER_TYPE, OTHER_SUBTYPE, NULL, zm_start_part, zm_add_part_data, zm_end_part);
		zm_part_handler_extract_data(&part_handler, msg, msg_len, part, ZM_EXTRACT_ENVELOPE);
		zm_part_handler_destroy(&part_handler);

		zm_address_parser_destroy(&address_parser);
		zm_iconv_destroy(&iconv);
		zm_data_buffer_destroy(&buffer);
	}

	zm_free_part(part);
	if (argc == 3) { zm_free_aliases(aliases); }

	munmap(msg, msg_len);
	close(fh);
return 0;
}

