/*-
 *  Copyright (c) Iris Dev. team. All rights reserved.
 *  See http://www.communico.pro/products/zeromime/license/ for details.
 *
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <strings.h>
#include <sysexits.h>
#include <unistd.h>

#include "zm_aliases.h"
#include "zm_block.h"
#include "zm_data_buffer_impl.h"
#include "zm_mem.h"
#include "zm_parser.h"
#include "zm_struct.h"

static int32_t level = 0;

static void store_part_offset(ZM_DATA_BUFFER  * dst)
{
	int32_t pos = 1;

	for (; pos < level; ++pos)
	{
		zm_data_buffer_append_data(dst, "    ", 4);
	}
}

static void print_part_offset()
{
	int32_t pos = 0;

	for (; pos < level; ++pos)
	{
		fprintf(stdout, "    ");
	}
}

static void print_part_name_rec(ZM_PART  * part)
{
	if (part == NULL) { return; }

	print_part_name_rec(part -> parent);
	if (part -> parent != NULL && ! part -> is_text) { fprintf(stdout, "%u.", part -> part_num); }
}

static void print_part_name(ZM_PART  * part)
{
	if (part == NULL) { return; }

	fprintf(stdout, "Part: ");

	print_part_name_rec(part -> parent);
	if (part -> is_text)
	{
		fprintf(stdout, "TEXT");
	}
	else
	{
		fprintf(stdout, "%u", part -> part_num);
	}

	fprintf(stdout, " ");
}

static void print_part_data(void            * msg,
                            ZM_PART         * part,
                            ZM_DATA_BUFFER  * dst)
{
	ZM_FIELD_REC  * field = NULL;

	uint64_t      type_len;
	const char  * type    = zm_get_part_type(part, msg, &type_len);

	uint64_t      subtype_len;
	const char  * subtype = zm_get_part_subtype(part, msg, &subtype_len);

	store_part_offset(dst);
	zm_data_buffer_append_data(dst, type,    type_len);
	zm_data_buffer_append_data(dst, "/",     1);
	zm_data_buffer_append_data(dst, subtype, subtype_len);
	zm_data_buffer_append_data(dst, "\n",    1);

	print_part_offset();
	print_part_name(part);

	if (part -> type == TEXT)
	{
		uint64_t      charset_len;
		const char  * charset = zm_get_text_part_charset(&(part -> u.text), msg, &charset_len);

		fprintf(stdout, "%.*s/%.*s; charset=%.*s (%u)\n", (int32_t)type_len, type, (int32_t)subtype_len, subtype, (uint32_t)charset_len, charset, (uint32_t)(part -> u.text.charset));
	}
	else
	{
		fprintf(stdout, "%.*s/%.*s\n", (int32_t)type_len, type, (int32_t)subtype_len, subtype);
	}

	field = zm_part_first_header(part);
	while (field != NULL)
	{
		ZM_KEY_VAL * key_val = zm_field_val_first(&(field -> value));

		uint64_t      name_len;
		const char  * name      = zm_block_get_data(&(field -> name), msg, &name_len);

		uint64_t      value_len;
		const char  * value     = zm_block_get_data(&(field -> value.raw), msg, &value_len);

		if(name_len > sizeof("Content-") && strncasecmp("Content-", name, sizeof("Content-") -1) == 0)
		{
			print_part_offset();
			fprintf(stdout, "`%.*s` => `%.*s`", (int32_t)name_len, name, (int32_t)value_len, value);

			while (key_val != NULL)
			{
				uint64_t      key_len;
				const char  * key     = zm_block_get_data(&(key_val -> key), msg, &key_len);

				uint64_t      val_len;
				const char  * val     = zm_block_get_data(&(key_val -> val), msg, &val_len);

				fprintf(stdout, "; `%.*s` => `%.*s`", (int32_t)key_len, key, (int32_t)val_len, val);

				key_val = zm_key_val_next(key_val);
			}
			fprintf(stdout, "\n");
		}

		field = zm_part_header_next(field);
	}
	fprintf(stdout, "\n");
}

static void dump_part_data(void            * msg,
                           ZM_PART         * part,
                           ZM_DATA_BUFFER  * dst)
{
	++level;
	print_part_data(msg, part, dst);

	if (part -> type == MULTIPART)
	{
		ZM_PART  * mpart = zm_multipart_first_subpart(part);
		while (mpart != NULL)
		{
			dump_part_data(msg, mpart, dst);
			mpart = zm_next_part(mpart);
		}
	}
	else if(part -> type == MESSAGE)
	{
		dump_part_data(msg, zm_message_subpart(part), dst);
	}

	--level;
}

static int32_t compare_part_structure(ZM_DATA_BUFFER  * dst,
                                      const char      * s_file)
{
	int32_t           fh       = 0;
	const char      * file     = NULL;
	uint32_t          file_len = 0;
	struct stat       st;
	const char      * dst_str = zm_data_buffer_get(dst);
	uint32_t          dst_len = zm_data_buffer_get_size(dst);
	uint32_t          pos = 0;
	int32_t           rc  = 0;

	fh = open(s_file, O_NONBLOCK, O_RDONLY);
	if (fh == -1)
	{
		fprintf(stderr, "Can't open file `%s`\n", s_file);
		return -1;
	}

	if (fstat(fh, &st) == -1)
	{
		fprintf(stderr, "Can't stat file `%s`\n", s_file);
		close(fh);
		return -1;
	}

	file_len = st.st_size;
	if (dst_len != file_len)
	{
		fprintf(stderr, "Sizes differ %u != %u\n", dst_len, file_len);
		close(fh);
		return -1;
	}

	file = (const char *)mmap(NULL, file_len, PROT_READ, MAP_PRIVATE, fh, 0);
	if (file == MAP_FAILED)
	{
		fprintf(stderr, "Can't map file `%s` into memory\n", s_file);
		close(fh);
		return -1;
	}

	for(pos = 0; pos < dst_len; ++pos)
	{
		if (toupper(file[pos] ) != toupper(dst_str[pos]))
		{
			fprintf(stderr, "Found difference here:\n");
			fprintf(stderr, "===============>%.*s\n\n", dst_len - pos, dst_str + pos);
			fprintf(stderr, "===============>%.*s\n\n", dst_len - pos, file + pos);
			rc = -1;
			break;
		}
	}

	munmap((void *)file, file_len);
	close(fh);

return rc;
}

int main(int argc, char ** argv)
{
	ZM_PART         * part = NULL;
	ZM_ALIASES      * aliases = NULL;

	int32_t           fh      = 0;
	void            * msg     = NULL;
	uint64_t          msg_len = 0;

	ZM_DATA_BUFFER    dst;
	const char      * dst_str  = NULL;
	int32_t           dst_len = 0;
	int32_t           rc      = 0;
	int32_t           code    = EX_SOFTWARE;

	struct stat       st;

	if (argc != 3 && argc != 4) { fprintf(stderr, "usage: %s message.mime mime-structure-dump.txt [aliases.txt]\n", argv[0]); return EX_USAGE; }

	fh = open(argv[1], O_NONBLOCK, O_RDONLY);
	if (fh == -1)
	{
		fprintf(stderr, "Can't open file `%s`\n", argv[1]);
		return EX_SOFTWARE;
	}

	if (fstat(fh, &st) == -1)
	{
		fprintf(stderr, "Can't stat file `%s`\n", argv[1]);
		close(fh);
		return EX_SOFTWARE;
	}

	msg_len = st.st_size;

	msg = mmap(NULL, msg_len, PROT_READ, MAP_PRIVATE, fh, 0);
	if (msg == MAP_FAILED)
	{
		fprintf(stderr, "Can't map file `%s` into memory\n", argv[1]);
		close(fh);
		return EX_SOFTWARE;
	}

	if (argc == 4)
	{
		if (zm_load_aliases(argv[3], &zm_default_allocator, &aliases) == -1) { return EX_SOFTWARE; }
	}

	zm_parse_mime(msg, msg_len, &zm_default_allocator, 1, aliases, &part);

	if (part != NULL)
	{
		zm_data_buffer_create_impl(&dst, &zm_default_allocator);

		code = EX_OK;
		dump_part_data(msg, part, &dst);

		rc = compare_part_structure(&dst, argv[2]);
		if (rc == 0)
		{
			code = EX_OK;
		}
		else
		{
			fprintf(stderr, "ERROR\n");
		}

		dst_str = zm_data_buffer_get(&dst);
		dst_len = zm_data_buffer_get_size(&dst);

		fprintf(stdout, "%.*s", dst_len, dst_str);

		zm_data_buffer_destroy(&dst);
	}

	zm_free_part(part);

	if (argc == 4) { zm_free_aliases(aliases); }

	munmap(msg, msg_len);
	close(fh);

return code;
}

